class SetKey():
    """
    This class sets a key as the index
    
    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    """
    def __init__(self, key):
        self.key = key

    def fit(self, X, y):
        return self
        
    def transform(self, X):
        key_in_cols = self.key in X.columns
        if key_in_cols:
            X = X.set_index(self.key)
        return X




import warnings
import pandas as pd
import numpy as np
warnings.filterwarnings(action="once")
import datetime as dt
import os
import sys
import re
sys.path.append("../")

from sklearn.preprocessing import MultiLabelBinarizer
from category_encoders import TargetEncoder

from sklearn.base import BaseEstimator, TransformerMixin 

from joblib import Parallel, delayed
from tqdm import tqdm

class AppFeatureEngineering(BaseEstimator, TransformerMixin):
    """This class is used to execute general ZRP feature engineering.
    
    Parameters
    ----------
    key: str 
        Key to set as index
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    geo_key: str
        Name of Census GEOID column
    race: str
        Name of race column
    """ 
    
    def __init__(self, key = "ZEST_KEY", geo_key = "GEOID", first_name ="first_name" , middle_name ="middle_name", last_name ="last_name", race = "race"):
        self.key = key
        self.geo_key = geo_key
        self.first_name = first_name
        self.middle_name = middle_name
        self.last_name = last_name
        self.race = race
        self.label_encoded_columns = [self.first_name, self.last_name, self.middle_name] 
        self.keys = [self.key, self.geo_key]

    def _process_target(self, y): 
        y_unique = y.unique()
        y_unique.sort()
        self.n_classes = len(y_unique)
        
        possible_race_classes = ["AAPI", "AIAN",  "BLACK", "HISPANIC", "WHITE"]
        
        # handle multi-labeled output
        self.mlb = MultiLabelBinarizer(classes = y_unique)
        self.mlb_columns = list(set(possible_race_classes) & set(y_unique))
        self.mlb.fit(y.values.reshape(-1,1))
        y_ohe = pd.DataFrame(self.mlb.transform(y.values.reshape(-1,1)), columns=self.mlb_columns)
        
        self.le = {}
        for i in range(self.n_classes):
            self.le[i] = TargetEncoder()
        return y_ohe
    
    def fit(self, X, y):
        targets = X[[self.key]].merge(y.reset_index(drop=False), on=self.key, how="left")
        y = targets[self.race]
#         y = targets.set_index(self.key)[self.race]
        X = X.reset_index(drop=True)
#         y = y.reset_index(drop=True) 

        self.data_columns = list(X.columns)
        self.acs_columns = list(set(self.data_columns) - set(self.label_encoded_columns) - set(self.keys))
        
        y_ohe = self._process_target(y)
        
        # fit label encoded columns
        for i in range(self.n_classes):
            self.le[i].fit(X[self.label_encoded_columns], y_ohe.iloc[:,i])
        
        return self
    
    def transform(self, X):
        X = X.reset_index(drop=False)
        data_fe = pd.concat([self.le[i].transform(X[self.label_encoded_columns]) for i in range(self.n_classes)],
                         axis=1, sort=False
                        )
        
        data_fe = pd.concat([data_fe,
                          X[self.keys],
                             X[self.acs_columns]
                         ], axis=1, sort=False)
        label_encoded_colname = []
        for label in self.mlb_columns:
            for col in self.label_encoded_columns:
                label_encoded_colname.append(label + "_" + col)
                
        data_fe.columns = label_encoded_colname +  self.keys + self.acs_columns
        data_fe[label_encoded_colname] = data_fe[label_encoded_colname].astype(float)
        
        return data_fe
    



      
    
    
class NameAggregation(BaseEstimator, TransformerMixin):
    """
    This class aggregates across expected name columns to impose one-to-one data
    
    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    n_jobs: int (default 1)
        Number of jobs in parallel
    """
    def __init__(self, key, n_jobs):
        self.key = key
        self.n_jobs =  n_jobs
        self.keys = [self.key]

    def fit(self, X, y):
        all_mto_feats = ['AAPI_first_name',
                         'AAPI_last_name',
                         'AAPI_middle_name',
                         'AIAN_first_name',
                         'AIAN_last_name',
                         'AIAN_middle_name',
                         'BLACK_first_name',
                         'BLACK_last_name',
                         'BLACK_middle_name',
                         'HISPANIC_first_name',
                         'HISPANIC_last_name',
                         'HISPANIC_middle_name',
                         'MULTIRACIAL_first_name',
                         'MULTIRACIAL_last_name',
                         'MULTIRACIAL_middle_name',
                         'OTHER_first_name',
                         'OTHER_last_name',
                         'OTHER_middle_name',
                         'WHITE_first_name',
                         'WHITE_last_name',
                         'WHITE_middle_name']
        self.mto_feats = list(set(X.columns).intersection(set(all_mto_feats)))
        self.drop_cols = self.mto_feats 
        return self
        
    def agg_col(self, X):
        da_agg = X.groupby(self.key).agg("mean")
        return(da_agg)
        
    def transform(self, X):
        df = X.copy()
        df = df.filter(self.mto_feats + [self.key])
        df = df.sort_values('ZEST_KEY')
        
        chunks = [df[x:x+10000] for x in range(0, len(df), 10000)]
        results = Parallel(n_jobs=self.n_jobs, verbose=1, prefer='threads')(delayed(self.agg_col)(chunk) for chunk in tqdm(chunks))
        
        aggd_data = pd.concat(results)  
        aggd_columns= list(aggd_data.columns)
        aggd_data = aggd_data[~aggd_data.index.duplicated(keep='first')]  
        aggd_data[aggd_columns] = aggd_data[aggd_columns].apply(lambda x: x.round(5))
        
        X = X.drop(self.drop_cols, axis=1)
        X = X.drop_duplicates().set_index(self.key)
        aggd_data = aggd_data.sort_index()
        X = X.sort_index()
        
        data_out = pd.merge(X, aggd_data, left_index=True, right_index=True)
                
        return(data_out)
    


from os.path import dirname, join, expanduser
from zrp.validate import *
from .preprocessing import *
from .geo_geocoder import *
from .acs_mapper import *
from .base import BaseZRP
from .utils import *
import pandas as pd
import numpy as np
import warnings
import glob
import json
import sys
import os
import re

import warnings
warnings.filterwarnings(action='ignore')

class ZRP_Prepare(BaseZRP):
    """
    Prepares data to generate race & ethnicity proxies

    Parameters
    ------------
    file_path: str, optional
        Path where to put artifacts and other files generated during intermediate steps.
    """
    
    def __init__(self, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)

        
    def fit(self, data):
        if self.census_tract:
            tract_lengths =  data[self.census_tract].str.len()
            tract_len  = most_common(tract_lengths)
            assert (data[self.census_tract].isalnum).any(), "Cannot provide non-numeric Census Tract code, please remove non-numeric census tract records "
            assert tract_len == 11,  "Improper Census Tract format provided. The tool requires the full state fips, county fips, and tract format. (ie '010010202001')"
        if self.block_group:
            bg_lengths =  data[self.block_group].str.len()
            bg_len  = most_common(bg_lengths)
            assert tract_len == 12,  "Improper Census Block Group format provided. The tool requires the full state fips, county fips, tract, and block group format. (ie '0100102020011')"            
    
    def transform(self, input_data):
        """
        Parameters
        ----------
        input_data: pd.Dataframe
            Dataframe to be transformed
        """  
        curpath = dirname(__file__)
        
        # Load Data
        try:
            data = input_data.copy()
            print("Data is loaded")
        except AttributeError:
            data = load_file(self.file_path)
            print("Data file is loaded")
            
        gen_process = ProcessStrings(file_path=self.file_path)
        gen_process.fit(data)
        data = gen_process.transform(data)
        
        processed = True
        replicate = True
        print("")

        print("[Start] Preparing geo data")
        data_path = join(curpath, f'../data/processed')
        assert len(os.listdir(data_path)) > 0, "Missing required support files please see the README for how to download the support files: https://github.com/zestai/zrp/blob/main/README.rst#install "        
        
        inv_state_map = load_json(join(data_path, "inv_state_mapping.json"))
        data['zest_in_state_fips'] = data[self.state].replace(inv_state_map)
        print("")

        if self.census_tract is not None:
            try:
                data[self.zip_code] = np.where((data[self.zip_code].isna()) |\
                                        (data[self.zip_code].str.contains("None")),
                                           None,
                                           data[self.zip_code].apply(lambda x: x.zfill(5)))
                data['GEOID_ZIP'] = data[self.zip_code]

            except (ValueError, KeyError) as e:
                pass
            geo_coded = data.copy()
            try:
                geo_coded['GEOID_BG'] = geo_coded[self.block_group]
            except (ValueError, KeyError) as e:
                pass
            geo_coded['GEOID_CT'] = geo_coded[self.census_tract]
            
        elif (self.census_tract is not None) & (self.street_address is not None):
            geocode = ZGeo(file_path=self.file_path)
            geocode.fit(geo_coded)
            geocode_out = [] 
            geo_grps = data.groupby([self.state])
            geo_dict = {}
            for s, g in geo_grps:
                geo_dict[s] = g
            gdkys = list(geo_dict.keys())
            print("  The following states are included in the data:", gdkys)
            
            assert set(gdkys) <= set(list(inv_state_map.keys())), "Provided non-standard state codes. Please use standard 2-letter abbreviation to indicate states to geocode, ex:'CA' for Californina"
            
            geo_out = [] 
            for s in tqdm(gdkys):                
                print(" ... on state:", str(s))
                geo = inv_state_map[s].zfill(2)
                output = geocode.transform(geo_dict[s], geo, True)
                geocode_out.append(output)
            geo_coded = pd.concat(geocode_out)
            geo_coded = geo_coded.drop_duplicates()
        else:
            geocode = ZGeo(file_path=self.file_path)
            geocode_out = [] 
            geo_grps = data.groupby([self.state])
            geo_dict = {}
            for s, g in geo_grps:
                geo_dict[s] = g
            gdkys = list(geo_dict.keys())
            print("  The following states are included in the data:", gdkys)
                  
            assert set(gdkys) <= set(list(inv_state_map.keys())), "Provided non-standard state codes. Please use standard 2-letter abbreviation to indicate states to geocode, ex:'CA' for Californina"

            geo_out = [] 
            for s in tqdm(gdkys):
                print("   ... on state:", str(s))
                geo = inv_state_map[s].zfill(2)
                output = geocode.transform(geo_dict[s], geo, processed, replicate, True)
                geocode_out.append(output)
            geo_coded = pd.concat(geocode_out)
        
        # append data unable to enter geo mapping
        geo_coded_keys = list(geo_coded.ZEST_KEY_COL.values)
        data = data[~data.index.isin(geo_coded_keys)]
        geo_coded = pd.concat([geo_coded, data])
        

        print("")
        
        print("[Completed] Preparing geo data")
        print("")
        print("[Start] Preparing ACS data")
        
        print("   [Start] Validating ACS input data")
        validate = ValidateGeocoded()
        validate.fit()
        acs_validator = validate.transform(geo_coded)
        save_json(acs_validator, self.out_path, "input_acs_validator.json")
        print("   [Completed] Validating ACS input data")
        print("")

        amp = ACSModelPrep()
        amp.fit()
        data_out = amp.transform(geo_coded, False)
        print("[Complete] Preparing ACS data")
        print("")
        
        return(data_out)



import pandas as pd
import numpy as np


def is_missing(data, required_cols):
    """Checks if all required columns are provided
    Parameter
    ---------
    data: pd.dataframe
        dataframe to make changes to or use for validation
    """
    missing_columns = np.setdiff1d(required_cols,data.columns).tolist()
    return(missing_columns)

class BaseValidate():
    """
    This is a base validator which all validators inherit from.
    
    Parameters
    ----------
    data: dataframe
        dataframe with user data
    state_mapping: dictionary
        dictionary mapping state names & abbreviations
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    block_group: str
        Name of block group column        
    race: str
        Name of race column
    year: str
        Name of year column
    span: str
        Name of span column
    """


    
    def __init__(self, support_files_path = "data/processed", key = "ZEST_KEY", first_name = "first_name", middle_name = "middle_name", last_name = "last_name", house_number = "house_number", street_address = "street_address", city = "city", state = "state", zip_code = "zip_code", race = "race", census_tract =  None, block_group = None, file_path = None, year = "2019", span  = "5"):
        self.key = key
        self.first_name = first_name
        self.middle_name =  middle_name
        self.last_name = last_name
        self.house_number = house_number
        self.street_address = street_address
        self.city = city
        self.state = state
        self.zip_code = zip_code
        self.census_tract = census_tract
        self.block_group = block_group
        self.file_path = file_path
        self.race =  race
        self.year =  year
        self.span = span
        
    def fit(self, data):  
        data_cols = data.columns
        self.names = list(set(data_cols).intersection([self.first_name, self.middle_name, self.last_name]))
        self.geo_numerics = list(set(data_cols).intersection([self.zip_code, self.census_tract, self.block_group, self.house_number]))
        self.geo_strings = list(set(data_cols).intersection([self.street_address, self.city, self.state]))

        return self
        
        
    def check_row_counts(self, data):
        """Calculate row counts
        
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        return(len(data))
        
    def check_missing_pct(self, data, is_input = True):
        """Calculates percentage of missing values
        
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        is_input: bool
            Indicator if validating raw input data
        """
        data_cols = list(data.columns)
        possible_zrp_cols =  set(data_cols).intersection(set([
            self.first_name, self.middle_name, self.last_name,
            self.house_number, self.street_address, self.zip_code,
            self.state, self.block_group, self.census_tract,
        ]))
        na_dict = {}
        for col in data_cols:
            na_dict[col] = None
            na_dict[col] = data[(data[col].astype(str).str.upper() == "NONE")
                                | (data[col].astype(str).str.upper() == " ")
                                | (data[col].isna())].shape[0]/data.shape[0]
            if col in possible_zrp_cols:
                if na_dict[col] > 0.10:
                    print(f"       (Warning!!) {col} is {na_dict[col]*100}% missing, this may impact the ability to return race approximations")   
        return(na_dict)
    
    def is_geocoded(self, data):
        """Calculates how much data is geocoded by geo-level
        
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        geocoded_cts = {}
        geocoded_cts["count"] = {}
        
        geocoded_cts["count"]["GEOID"] = data[(data["GEOID"].str.len()>4)
                                                       & (data.index.duplicated(keep = "first"))].shape[0]
        geocoded_cts["count"]["Block Group"] = data[(data["GEOID_BG"].str.len()>11)  
                                                    & (data["GEOID_BG"] == data["GEOID"])
                                                    & (data["GEOID_BG"].notna())].shape[0]
        geocoded_cts["count"]["Census Tract"] = data[(data["GEOID_CT"].str.len()>10) 
                                                     & (data["GEOID_CT"] == data["GEOID"])
                                                     & (data["GEOID_CT"].notna())].shape[0]
        geocoded_cts["count"]["Zip Code"] = data[(data["GEOID_ZIP"].str.len() == 5)  
                                                 & (data["GEOID_ZIP"] == data["GEOID"])
                                                 & (data["GEOID_ZIP"].notna())].shape[0]
        return(geocoded_cts)
        
    def check_states(self, data):
        """Count of all states in user input data
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        return(data[self.state].value_counts(dropna = False).to_dict())
    
    def is_zcta5(self, data):
        """Determines if zip codes are provided as 5 digit zip codes
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        geo_dict = {}
        geo_dict["length_check"] = data[self.zip_code].str.len().value_counts(dropna = False).to_dict()
        geo_dict["numeric_check"] = data[self.zip_code].str.isnumeric().all()
        return(geo_dict)
    
    def is_census_tract(self, data):
        """Determines if standard census tracts are provided
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        geo_dict = {}
        geo_dict["length_check"] = data[self.census_tract].str.len().value_counts(dropna = False).to_dict()
        geo_dict["numeric_check"] = data[self.census_tract].str.isnumeric().all()
        return(geo_dict)
    
    def is_block_group(self, data):
        """Determines if standard block groups are provided
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        geo_dict = {}
        geo_dict["length_check"] = data[self.block_group].str.len().value_counts(dropna = False).to_dict()
        geo_dict["numeric_check"] = data[self.block_group].str.isnumeric().all()
        return(geo_dict)
    
    def is_mapped(self, data):
        """Determines how much data is mapped
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        mapped_dict = {}
        for acssrc in ["BG", "CT", "ZIP"]:
            data["acs_source"]
            mapped_sum = data[data["acs_source"] == acssrc].sum()
            mapped_dict [acssrc] = None
            mapped_dict[acssrc]["count"] = mapped_sum
            if mapped_sum>0:
                pct_mapped = mapped_sum/data.shape[0]
                mapped = True
            else:
                mapped_dict[acssrc]["percent"] = 0
        return(mapped_dict)
            
    def is_empty(self, data):
        """Checks if dataframe is empty
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        return(bool(data.empty))
    
    def is_all_missing(self, data):
        """Checks if all data in dataframe is missing
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        return(bool(data.isna().all().all()))

    def is_unique_key(self, data):
        """Checks if the provided key is unique
                
        Parameter
        ---------
        data: pd.dataframe
            dataframe to make changes to or use for validation
        """
        if self.key in data.columns:
            out = data[self.key].nunique() == data.shape[0]
        else:
            out = data.index.nunique() == data.shape[0]
        return(bool(out))
 
            
    def transform(self, data):
        pass
    

class ValidateGeo(BaseValidate):
    """
    Validates geo data
    
    Parameters
    ----------
    data: dataframe
        dataframe with user data
    state_mapping: dictionary
        dictionary mapping state names & abbreviations
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    block_group: str
        Name of block group column        
    race: str
        Name of race column
    year: str
        Name of year column
    span: str
        Name of span column
    """    
        
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def fit(self):
        return self
            
    def transform(self, data):    
        validator = {}
        if self.zip_code in data.columns:
            validator["is_zip_code"] = self.is_zcta5(data)
        if self.is_census_tract in data.columns:
            validator["is_census_tract"] = self.is_census_tract(data)
        if self.is_block_group in data.columns:
            validator["is_block_group"] = self.is_block_group(data)
        return(validator)
        

class ValidateInput(BaseValidate):
    """
    Validates user input data
    
    Parameters
    ----------
    data: dataframe
        dataframe with user data
    state_mapping: dictionary
        dictionary mapping state names & abbreviations
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    block_group: str
        Name of block group column        
    race: str
        Name of race column
    year: str
        Name of year column
    span: str
        Name of span column
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def fit(self):
        return self
            
    def transform(self, data):    
        validator = {}
        validator["is_empty"] = self.is_empty(data)
        if validator["is_empty"]:
            raise ValueError("Dataframe is empty")
        validator["is_all_missing"] = self.is_all_missing(data)
        if validator["is_all_missing"]:
            raise ValueError("Dataframe is fully missing")
        validator["n_obs"] = self.check_row_counts(data)
        print("     Number of observations:", validator["n_obs"])
        validator["is_unique_key"] = self.is_unique_key(data)
        print("     Is key unique:", validator["is_unique_key"]) 
        validator["pct_na"] = self.check_missing_pct(data)
        return(validator)
            
class ValidateGeocoded(BaseValidate):
    """
    Validates input ACS data that has geocoded keys.
    
    Parameters
    ----------
    data: dataframe
        dataframe with user data
    state_mapping: dictionary
        dictionary mapping state names & abbreviations
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    block_group: str
        Name of block group column        
    race: str
        Name of race column
    year: str
        Name of year column
    span: str
        Name of span column
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def fit(self):
        return self
            
    def transform(self, data):    
        validator = {}
        validator["is_empty"] = self.is_empty(data)
        if validator["is_empty"]:
            raise ValueError("Dataframe is empty")
        validator["is_all_missing"] = self.is_all_missing(data)
        if validator["is_all_missing"]:
            raise ValueError("Dataframe is fully missing")
        validator["n_obs"] = self.check_row_counts(data)
        print("     Number of observations:", validator["n_obs"])
        validator["is_unique_key"] = self.is_unique_key(data)
        print("     Is key unique:", validator["is_unique_key"]) 
                
        validator["pct_na"] = self.check_missing_pct(data, is_input=False)
        print("")
        try:
            for i in [self.last_name, self.first_name]:
                tmp = round(validator["pct_na"][i]*100, 2)
                assert tmp < 10, f"Too many missing values in required name feature, {i}. {tmp}% of the values are missing. Please review data and reduce missing. Required features include first name and last name." 
        except (KeyError, ValueError) as e:
            pass
        validator["is_geocoded"] = self.is_geocoded(data)
        return(validator)        
    
    
class ValidateACS(BaseValidate):
    """
    Validates ACS data post mapping
    
    Parameters
    ----------
    data: dataframe
        dataframe with user data
    state_mapping: dictionary
        dictionary mapping state names & abbreviations
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    block_group: str
        Name of block group column        
    race: str
        Name of race column
    year: str
        Name of year column
    span: str
        Name of span column
    """
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def fit(self):
        return self
            
    def transform(self, data):    
        validator = {}
        validator["is_empty"] = self.is_empty(data)
        if validator["is_empty"]:
            raise ValueError("Dataframe is empty")
        validator["is_all_missing"] = self.is_all_missing(data)
        if validator["is_all_missing"]:
            raise ValueError("Dataframe is fully missing")
        validator["n_obs"] = self.check_row_counts(data)
        print("     Number of observations:", validator["n_obs"])
        validator["is_unique_key"] = self.is_unique_key(data)
        print("     Is key unique:", validator["is_unique_key"]) 
        validator["is_mapped"] = self.is_mapped(data)
        return(validator)    


from .preprocessing import *
from os.path import dirname, join, expanduser
from joblib import Parallel, delayed
from .utils import *
from tqdm import tqdm
import pandas as pd
import numpy as np
import censusdata
import json
import sys
import os
import re


def pad_logrecno(data):
    data["LOGRECNO"] = data["LOGRECNO"].str.zfill(7)
    return (data)


class ACS_Parser():
    """
    Parses the raw ACS data
    
    Parameters:
    ----------
    support_files_path: str
        Support files path pointing to where the raw ACS data is stored
    year: str (default 2019)
        Year of ACS data. 
    span: str (default 5)
        Span of ACS data. The ACS data is available in 1 or 5 year spans. The 5yr ACS data is the most comprehensive & is available at more granular levels than 1yr data.
    state_level: str
        State to parse
    n_jobs: int
        Number of jobs in parallel        
    """
    
    def __init__(self, support_files_path, state_level, n_jobs=-1, year='2019', span='5'):
        self.support_files_path = support_files_path
        self.year = year
        self.span = span
        self.state_level = state_level
        self.n_jobs = n_jobs
        self.raw_acs_path = os.path.join(self.support_files_path,
                                         f"raw/acs/{self.year}/{self.span}yr")
        self.out_acs_path = os.path.join(self.support_files_path,
                                         "parsed/acs",
                                         self.year,
                                         f"{self.span}yr")
        self.fol_acs_tmp = f"{self.year}_{self.span}yr_Summary_FileTemplates/"

    def fit(self):
        return self

    def acs_geo(self, data):
        """
        Merges input data with ACS data
        
        Parameters:
        ----------
        data: dataframe
            Original input dataframe
        """
        geo_file_path = os.path.join(self.raw_acs_path,
                                     self.fol_acs_tmp,
                                     f'{self.span}_year_Mini_Geo.xlsx')

        ag_df = pd.read_excel(geo_file_path,
                              sheet_name=self.state_level,
                              dtype=str)
        ag_df = ag_df.rename(columns={'Logical Record Number': 'LOGRECNO',
                                      'Geography ID': 'GEOID'})

        data = data.merge(ag_df,
                          on=["LOGRECNO"])
        return (data)

    def acs_track_5yr(self):
        """
        Initializes 5 year sequence dictionary
            
        """
        file_path_list = []
        sequence_dict = {}
        for root, dirs, files in os.walk(os.path.join(self.raw_acs_path,
                                                      "Tracts_Block_Groups")):
            for file in files:
                if ('000.txt' in file) & \
                        ('e' in file) & \
                        (f'{self.span}{self.state_level}' in file):
                    file_path_list.append("".join([root, file]))
                    sequence = file.split('000.txt')[0].split(self.state_level)[1]
                    sequence = str(int(sequence))
                    sequence_dict[sequence] = {}

                    sequence_dict[sequence]["data"] = None
                    sequence_dict[sequence]["sequence"] = None
                    sequence_dict[sequence]["headers"] = None

                    sequence_dict[sequence]['Tracts_Block_Groups'] = {}
                    sequence_dict[sequence]['Not_Tracts_Block_Groups'] = {}

                    sequence_dict[sequence]['Tracts_Block_Groups']["file"] = file
                    sequence_dict[sequence]['Not_Tracts_Block_Groups']["file"] = None

                    sequence_dict[sequence]['Tracts_Block_Groups']["data"] = None
                    sequence_dict[sequence]['Not_Tracts_Block_Groups']["data"] = None

        for root, dirs, files in os.walk(os.path.join(self.raw_acs_path, "Not_Tracts_Block_Groups")):
            for file in files:
                if ('000.txt' in file) & ('e' in file) & \
                        (f'{self.span}{self.state_level}' in file):
                    file_path_list.append("".join([root, file]))
                    sequence = file.split('000.txt')[0].split(self.state_level)[1]
                    sequence = str(int(sequence))
                    sequence_dict[sequence]['Not_Tracts_Block_Groups']["file"] = file
        return (sequence_dict)

    def acs_track_1yr(self):
        """
        Initializes 1 year sequence dictionary
            
        """
        file_path_list = []
        sequence_dict = {}
        for root, dirs, files in os.walk(os.path.join(self.raw_acs_path,
                                                      "All_Geographies")):
            for file in files:
                if ('000.txt' in file) & \
                        ('e' in file) & \
                        (f'{self.span}{self.state_level}' in file):
                    file_path_list.append("".join([root, file]))
                    sequence = file.split('000.txt')[0].split(self.state_level)[1]
                    sequence = str(int(sequence))
                    sequence_dict[sequence] = {}
                    sequence_dict[sequence]["file"] = file
                    sequence_dict[sequence]["sequence"] = None
                    sequence_dict[sequence]["data"] = None
                    sequence_dict[sequence]["headers"] = None

        return (sequence_dict)

    def acs_parse_5yr(self, sequence_dict, i, save_table):
        """
        Parses 5 year ACS data
        
        Parameters:
        ----------
        sequence_dict: dict
            State level dictionary that contains file paths & sequence names
        i: int
            Sequence iteration
        save_table: bool
            Indicating is table will be saved to file
            
        """
        print("... Generating ACS table", self.state_level, "sequence", i)

        dat_file_1 = os.path.join(self.raw_acs_path,
                                  "Tracts_Block_Groups",
                                  sequence_dict[str(i)]['Tracts_Block_Groups']['file'])
        dat_file_2 = os.path.join(self.raw_acs_path,
                                  "Not_Tracts_Block_Groups",
                                  sequence_dict[str(i)]['Not_Tracts_Block_Groups']['file'])

        try:
            tmp_data_1 = pd.read_csv(dat_file_1, sep=",", header=None, dtype=str)
            tmp_data_2 = pd.read_csv(dat_file_2, sep=",", header=None, dtype=str)

            seq_file = os.path.join(self.raw_acs_path,
                                    sequence_dict[str(i)]['sequence'])

            tmp_headers = pd.read_excel(seq_file, sheet_name='e', dtype=str)
            sequence_dict[str(i)]["headers"] = tmp_headers

            feature_mapping = sequence_dict[str(i)]['headers'].to_dict('records')
            sequence_dict[str(i)]['description'] = feature_mapping[0]

            new_col_names = list(tmp_headers.columns)
            tmp_data_1.columns = new_col_names
            tmp_data_2.columns = new_col_names

        except pd.errors.EmptyDataError:
            print(f'   Note: {dat_file_1} was empty.')
            tmp_data = pd.read_csv(dat_file_2, sep=",", header=None, dtype=str)

            seq_file = os.path.join(self.raw_acs_path,
                                    sequence_dict[str(i)]['sequence'])
            tmp_headers = pd.read_excel(seq_file, sheet_name='e', dtype=str)

            sequence_dict[str(i)]["headers"] = tmp_headers

            feature_mapping = sequence_dict[str(i)]['headers'].to_dict('records')
            sequence_dict[str(i)]['description'] = feature_mapping[0]

            new_col_names = list(tmp_headers.columns)
            tmp_data.columns = new_col_names

        tmp_data = pad_logrecno(tmp_data)
        tmp_data = self.acs_geo(tmp_data)

        sequence_dict[str(i)]["data"] = tmp_data

        if save_table:
            file_name = "".join(["Zest_ACS_", self.state_level, "_seq",
                                 str(i), "_", self.year, '_', self.span,
                                 'yr.parquet'])
            save_dataframe(tmp_data, self.out_acs_path, file_name)
        return (sequence_dict[str(i)])

    def acs_parse_1yr(self, sequence_dict, i, save_table):
        """
        Parses 1 year ACS data
        
        Parameters:
        ----------
        sequence_dict: dict
            State level dictionary that contains file paths & sequence names
        i: int
            Sequence iteration
        save_table: bool
            Indicating is table will be saved to file
            
        """
        print("... Generating ACS table", self.state_level, "sequence", i)
        dat_file = os.path.join(self.raw_acs_path,
                                "All_Geographies",
                                sequence_dict[str(i)]['file'])

        tmp_data = pd.read_csv(dat_file, sep=",", header=None, dtype=str)
        seq_file = os.path.join(self.raw_acs_path,
                                sequence_dict[str(i)]['sequence'])
        tmp_headers = pd.read_excel(seq_file, sep=",",
                                    sheet_name="e", dtype=str)

        sequence_dict[str(i)]["headers"] = tmp_headers
        feature_mapping = sequence_dict[str(i)]['headers'].to_dict('records')
        sequence_dict[str(i)]['description'] = feature_mapping[0]

        new_col_names = list(tmp_headers.columns)
        tmp_data.columns = new_col_names
        tmp_data = pad_logrecno(tmp_data)
        tmp_data = self.acs_geo(tmp_data)

        sequence_dict[str(i)]["data"] = tmp_data

        if save_table:
            file_name = "".join(["Zest_ACS_", self.state_level, "_seq", str(i),
                                 "_", self.year, '_', self.span, 'yr.parquet'])

            save_dataframe(tmp_data, self.out_acs_path, file_name)
        return (sequence_dict[str(i)])

    def transform(self, save_table=True):
        """
        Parameters
        ----------
        save_table: bool
            Optional save
        """
        if save_table:

            make_directory(output_directory=self.raw_acs_path)
            make_directory(output_directory=self.out_acs_path)
            make_directory(output_directory=os.path.join(self.raw_acs_path,
                                                         self.fol_acs_tmp))
        if self.span == '5':
            sequence_dict = self.acs_track_5yr()
            for root, dirs, files in os.walk(os.path.join(self.raw_acs_path,
                                                          self.fol_acs_tmp)):
                for file in files:
                    if ('.xlsx' in file) & ('Geo' not in file):
                        sequence = file.split('seq')[1].split('.xlsx')[0]
                        sequence = str(int(sequence))
                        seq_str = "".join([self.fol_acs_tmp, "seq", sequence, ".xlsx"])
                        sequence_dict[sequence]["sequence"] = seq_str


            results = Parallel(n_jobs=self.n_jobs, verbose=1)(
                delayed((self.acs_parse_5yr))(sequence_dict, sni, save_table) for sni in
                tqdm(list(sequence_dict.keys())))

        elif self.span == '1':
            sequence_dict = self.acs_track_1yr()
            for root, dirs, files in os.walk(os.path.join(self.raw_acs_path,
                                                          self.fol_acs_tmp)):
                for file in files:
                    if ('.xlsx' in file) & ('Geo' not in file):
                        sequence = file.split('seq')[1].split('.xlsx')[0]
                        sequence = str(int(sequence))
                        seq_str = "".join([self.fol_acs_tmp, "seq", sequence, ".xlsx"])
                        sequence_dict[sequence]["sequence"] = seq_str


            results = Parallel(n_jobs=self.n_jobs, verbose=1)(
                delayed((self.acs_parse_1yr))(sequence_dict, sni, save_table) for sni in
                tqdm(list(sequence_dict.keys())))
        else:
            raise ValueError('Improper ACS span provided. The only accepted values are 1 & 5')
        results_out = {}
        for d in results:
            seqn = re.findall('[0-9]{1,3}', d['sequence'])[-1]
            results_out[seqn] = {}
            results_out[seqn].update(d)
        return (results_out)


def acs_census_data(support_files_path, level):
    """Create ACS Lookup Tables using the censusdata package.
    
    Parameters:
    -----------
    level:
        Geographic level to return ACS for, options include:
        ['block group' tract', 'zip', 'county', 'state']
    
    """
    curpath = dirname(__file__)
    data_path = join(curpath, '../data/')
    misc_args = load_json(join(data_path, 'misc_args.json'))
    states = load_json(join(data_path, 'states.json'))

    df = pd.DataFrame(columns=misc_args)
    if level == 'zip':
        temp = censusdata.download(
            "acs5",
            2019,
            censusdata.censusgeo([
                ('zip code tabulation area', '*')]),
            misc_args,
        )
        df = df.append(temp)

    else:
        pass

    for i in range(len(states)):
        print("State:", i, "\n")

        if level == 'tract':
            temp = censusdata.download(
                "acs5",
                2019,
                censusdata.censusgeo([
                    ('state', states[i]),
                    ('county', '*'),
                    ('tract', '*')
                ]),
                misc_args,
            )
            df = df.append(temp)


        elif level == 'block group':
            temp = censusdata.download(
                "acs5",
                2019,
                censusdata.censusgeo([
                    ('state', states[i]),
                    ('county', '*'),
                    ('block group', '*')
                ]),
                misc_args,
            )
            df = df.append(temp)

        else:
            temp = censusdata.download(
                "acs5",
                2019,
                censusdata.censusgeo([
                    ('state', states[i]),
                    ('county', '*')
                ]),
                misc_args,
            )
            df = df.append(temp)
    return (df)


class ACS_LookupBuilder():
    """Creates a core ACS lookup table by geo level 
    
    Parameter
    --------
    geo: str
        Geo key to identify which geographic level the ACS table will be made at. 
        Three levels are currently supported zip, tract, or block group
    year: str (default 2019)
        Year of ACS data. 
    span: str
        Span of ACS data. The ACS data is available in 1 or 5 year spans. The 5yr ACS data is the most comprehensive & is available at more granular levels than 1yr data.
    n_jobs: int
        Number of jobs in parallel
    required_tables: list
        List of ACS table names to select data to include in the Lookup table
    """


    def __init__(self, support_files_path, geo, year='2019', span='5', n_jobs=-1, required_tables=None):
        self.support_files_path = support_files_path
        self.geo = geo
        self.year = year
        self.span = span
        self.n_jobs = n_jobs
        self.support_files_path = support_files_path
        curpath = dirname(__file__)
        parsed_path = join(support_files_path, f'../data/parsed/acs/{self.year}/{self.span}yr')
        processed_path = join(curpath, f'../data/processed/acs/{self.year}/{self.span}yr')
        self.raw_acs_path = parsed_path
        self.out_acs_path = processed_path
        self.required_tables = required_tables
        if self.required_tables is None:
            self.required_tables = ['B01003', 'B02001', 'B03001', 'B04004', 'B04006',
                                    'B04007', 'B05011', 'B05012', 'B06009', 'B07101',
                                    'B08301', 'B10051', 'B11017', 'B16001', 'B19001',
                                    'B23020', 'B25004', 'B25075', 'B99021', 'B99162', 'C16001']

    def acs_select_features(self, data):
        """Selects predefined ACS tables to create a lookup table"""
        prts = "|".join(self.required_tables + ['GEO', 'Geography'])
        data = data.filter(regex=prts)
        return (data)

    def acs_join_lookup(self, data, mco):
        """Appends new geo level to feature column"""
        tbl_dict = {}
        for rn in required_names:
            data = data.filter(regex="|".join([rn, 'GEO']))
            tbl_dict[rn] = data
        return (tbl_dict)

    def parsed_acs_proc(self, file, geo_pattern):
        """Creates a dataframe of geo-specific ACS data by sequence 

        Parameter
        --------
        file: str
            Filename
        geo_pattern: str
            Pattern to identify geo specific data
        """
        drop_cols = []

        long_name = 'GEO_NAME'
        long_id = 'EXT_GEOID'

        tmp_data = load_file(os.path.join(self.raw_acs_path, file))
        tmp_data = self.acs_select_features(tmp_data)

        tmp_data = tmp_data.rename(columns={'GEOID': long_id,
                                            'Geography Name': long_name})

        tmp_data = tmp_data[tmp_data[long_name].str.upper().str.contains(geo_pattern,
                                                                         regex=True)]
        na_cols = list(tmp_data.columns[tmp_data.isnull().all()])
        tmp_data = tmp_data.drop(drop_cols + na_cols,
                                 axis=1)

        tmp_data['GEOID'] = None
        tmp_data['GEOID'] = tmp_data[long_id].apply(lambda x: x.split('US')[1]).astype(str)

        if tmp_data.shape[1] < 4:
            tmp_data = pd.DataFrame()
        else:
            tmp_data = tmp_data.set_index(['GEOID',
                                           'GEO_NAME',
                                           'EXT_GEOID']
                                          ).sort_index()
        return (tmp_data)

    def transform(self, save_table):
        """
        Parameters
        ----------
        save_table: bool
            Optional save
        """
        assert self.geo in ['zip', 'tract',
                            'block group'], "Require `geo` to be specified as 'zip', 'tract' or 'block group'"
        output = []

        if self.geo == 'zip':
            geo_pattern = '^ZCTA5'

            parsed_file_list = [file for file in os.listdir(self.raw_acs_path) if 'ACS_us_seq' in file]

            output = Parallel(n_jobs=self.n_jobs, verbose=10)(
                delayed(self.parsed_acs_proc)(file, geo_pattern) for file in parsed_file_list)

        elif self.geo == 'block group':
            geo_pattern = '^BLOCK GROUP'

            parsed_file_list = [file for file in os.listdir(self.raw_acs_path) if 'ACS_us_seq' not in file]
            results = Parallel(n_jobs=self.n_jobs, verbose=1)(
                delayed(self.parsed_acs_proc)(file, geo_pattern) for file in parsed_file_list)

            tbl_dict = {}
            for rn in self.required_tables:
                tbl_dict[rn] = {}
                tbl_dict[rn]['tbls'] = []
                tbl_dict[rn]['data'] = pd.DataFrame()

            for dfd in tqdm(range(0, len(results))):
                if not results[dfd].empty:
                    mco = most_common([x.split('_')[0] for x in results[dfd].columns])[0:6]
                    tbl_dict[mco]['tbls'].append(results[dfd])

            for rn in self.required_tables:
                print("  Table:", rn)
                try:
                    tbl_dict[rn]['data'] = pd.concat(tbl_dict[rn]['tbls'])
                    print("    ", tbl_dict[rn]['data'].shape)
                    output.append(tbl_dict[rn]['data'])
                except ValueError:
                    pass
            print("")

        elif self.geo == 'tract':
            geo_pattern = '^CENSUS TRACT'

            parsed_file_list = [file for file in os.listdir(self.raw_acs_path) if 'ACS_us_seq' not in file]
            results = Parallel(n_jobs=self.n_jobs, verbose=10)(
                delayed(self.parsed_acs_proc)(file, geo_pattern) for file in parsed_file_list)

            tbl_dict = {}
            for rn in self.required_tables:
                tbl_dict[rn] = {}
                tbl_dict[rn]['tbls'] = []
                tbl_dict[rn]['data'] = pd.DataFrame()

            for dfd in tqdm(range(0, len(results))):
                if not results[dfd].empty:
                    mco = most_common([x.split('_')[0] for x in results[dfd].columns])[0:6]
                    tbl_dict[mco]['tbls'].append(results[dfd])

            for rn in self.required_tables:
                print("  Table:", rn)
                try:
                    tbl_dict[rn]['data'] = pd.concat(tbl_dict[rn]['tbls'])
                    print("    ", tbl_dict[rn]['data'].shape)
                    output.append(tbl_dict[rn]['data'])
                except ValueError:
                    pass
            print("")

        df_out = pd.concat(output, axis=1)

        # optional save 
        if save_table:
            make_directory(output_directory = self.raw_acs_path)
            make_directory(output_directory = self.out_acs_path)
            geo_name = "".join(self.geo.split())
            filename = "".join(["Zest_ACS_Lookup_", self.year, self.span, 'yr_', geo_name, '.parquet'])
            print(f"Saving {filename}")
            save_dataframe(df_out, self.out_acs_path, filename)
        else:
            print('No tables were saved')

        return (df_out)



from os.path import expanduser, dirname
import pandas as pd
import numpy as np
import os
import re
import sys
import json
import joblib
import pickle

from xgboost import XGBClassifier
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.model_selection import train_test_split
from feature_engine.imputation import MeanMedianImputer
from feature_engine.selection import SmartCorrelatedSelection, DropFeatures

from zrp.modeling.src.app_preprocessor import HandleCompoundNames
from zrp.modeling.src.acs_scaler import CustomRatios
from zrp.modeling.src.app_fe import AppFeatureEngineering, NameAggregation
from zrp.modeling.src.set_key import SetKey
from zrp.prepare.utils import load_json, load_file, save_feather, make_directory
from zrp.prepare.base import BaseZRP
from zrp.prepare.prepare import ZRP_Prepare

import warnings
warnings.filterwarnings(action='ignore')

curpath = dirname(__file__)


class ZRP_Build_Pipeline(BaseZRP):
    """
    Fits a new ZRP pipeline from user input 
    
    Parameters
    ----------
    file_path: str, optional
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model
    zrp_model_source: str
        Indicates the source of zrp_modeling data to use. There are three optional sources 'bg' (for block_group), 'ct' (for census_tract), and 'zp' (for zip_code). By default 'census_tract' is inferred.
    """

    def __init__(self, file_path=None, zrp_model_name='zrp_0', zrp_model_source='ct', *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.zrp_model_name = zrp_model_name
        self.zrp_model_source = zrp_model_source
        self.outputs_path = os.path.join(self.out_path,
                                         "experiments",
                                         self.zrp_model_name,
                                         self.zrp_model_source,
                                         "data")
        self.geo_key = 'GEOID'

    def fit(self, X, y):
        ### Build Pipeline
        print('\n---\nBuilding pipeline')

        self.pipe = Pipeline(
            [("Drop Features", DropFeatures(features_to_drop=['GEOID_BG', 'GEOID_CT', 'GEOID_ZIP', 'ZEST_KEY_COL'])),
             ("Compound Name FE",
              HandleCompoundNames(last_name=self.last_name, first_name=self.first_name, middle_name=self.middle_name)),
             ("App FE", AppFeatureEngineering(key=self.key, geo_key=self.geo_key, first_name=self.first_name,
                                              middle_name=self.middle_name, last_name=self.last_name, race=self.race)),
             ("ACS FE", CustomRatios()),
             ("Name Aggregation", NameAggregation(key=self.key, n_jobs=self.n_jobs)),
             ("Drop Features (2)", DropFeatures(features_to_drop=['GEOID'])),
             ("Impute", MeanMedianImputer(imputation_method="mean", variables=None)),
             ("Correlated Feature Selection", SmartCorrelatedSelection(method='pearson',
                                                                       threshold=.95))],
            verbose=True
        )
        #### Fit the Pipeline
        print('\n---\nFitting pipeline')
        self.pipe.fit(X, y[self.race])

        return self

    def transform(self, X):
        make_directory(self.outputs_path)
        # Save pipeline
        pickle.dump(self.pipe, open(os.path.join(self.outputs_path, 'pipe.pkl'), 'wb'))
        #### Transform
        ##### This step creates the feature engineering data
        print('\n---\nTransforming FE data')
        X_train_fe = self.pipe.transform(X=X)

        # Save train fe data
        print('\n---\nSaving FE data')
        save_feather(X_train_fe, self.outputs_path, f"train_fe_data.feather")
        return (X_train_fe)


class ZRP_Build_Model(BaseZRP):
    """
    Generate as ZRP model from input data & pre-trained pipeline. 
    
    Parameters
    ----------
    file_path: str, optional
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model
    zrp_model_source: str
        Indicates the source of zrp_modeling data to use. There are three optional sources 'bg' (for block_group), 'ct' (for census_tract), and 'zp' (for zip_code). By default 'census_tract' is inferred.
    """

    def __init__(self, file_path=None, zrp_model_name='zrp_0', zrp_model_source='ct', *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.zrp_model_name = zrp_model_name
        self.zrp_model_source = zrp_model_source
        self.outputs_path = os.path.join(self.out_path,
                                         "experiments",
                                         self.zrp_model_name,
                                         self.zrp_model_source,
                                         "data")
        self.geo_key = 'GEOID'

    def fit(self, X, y):
        ### Build the zrp_model
        ##### specify zrp_model parameters
        print('\n---\nbuilding zrp_model')

        opt_params = {'gamma': 5,
                      'learning_rate': 0.01,
                      'max_depth': 3,
                      'min_child_weight': 500,
                      'n_estimators': 2000,
                      'subsample': 0.20}

        ##### Initialize the zrp_model
        self.zrp_model = XGBClassifier(objective='multi:softprob',
                                       num_class=len(y[self.race].unique()),
                                       **opt_params)
        ##### Fit
        print('\n---\nfitting zrp_model')
        self.zrp_model.fit(
            X, y[self.race],
            sample_weight=y.sample_weight
        )

        self.y_unique = y[self.race].unique()
        self.y_unique.sort()

        return self

    def transform(self, X):
        make_directory(self.outputs_path)
        # Save zrp_model
        pickle.dump(self.zrp_model, open(os.path.join(self.outputs_path, "zrp_model.pkl"), "wb"))

        ##### Return Race Probabilities
        print('\n---\nGenerate & save race predictions (labels)')
        y_hat_train = pd.DataFrame({'race': self.zrp_model.predict(X)}, index=X.index)

        y_hat_train.reset_index(drop=False).to_feather(os.path.join(self.outputs_path, f"train_proxies.feather"))

        print('\n---\nGenerate & save race predictions (probabilities)')
        y_phat_train = pd.DataFrame(self.zrp_model.predict_proba(X), index=X.index)

        #         y_phat_train.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        y_phat_train.columns = self.y_unique

        y_phat_train.reset_index(drop=False).to_feather(os.path.join(self.outputs_path, f"train_proxy_probs.feather"))

        return (y_hat_train, y_phat_train)


class ZRP_DataSampling(BaseZRP):
    """
    Generate data splits from input data 
    
    Parameters
    ----------
    file_path: str, optional
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model
    zrp_model_source: str
        Indicates the source of zrp_modeling data to use. There are three optional sources 'bg' (for block_group), 'ct' (for census_tract), and 'zp' (for zip_code). By default 'census_tract' is inferred.
    """

    def __init__(self, file_path=None, zrp_model_name='zrp_0', zrp_model_source='ct', *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.zrp_model_name = zrp_model_name
        self.zrp_model_source = zrp_model_source
        self.outputs_path = os.path.join(self.out_path,
                                         "experiments",
                                         self.zrp_model_name,
                                         self.zrp_model_source,
                                         "data")
        self.geo_key = 'GEOID'

    def fit(self):
        return self

    def transform(self, data):
        make_directory(self.outputs_path)
        df = data.copy()

        # Keep geocoded data & data with labels
        #         df = df[(df[self.geo_key].notna()) & (df[self.geo_key]!="None")]
        df = df[(df[self.race].notna()) & (df[self.race] != "None")]
        df_keys = list(df.index.unique())

        # sample weights normalizing to us population
        aapi_ratio = df[self.race].value_counts(normalize=True).AAPI if 'AAPI' in df[self.race] else 1
        black_ratio = df[self.race].value_counts(normalize=True).BLACK if 'BLACK' in df[self.race] else 1
        hispanic_ratio = df[self.race].value_counts(normalize=True).HISPANIC if 'HISPANIC' in df[self.race] else 1
        aian_ratio = df[self.race].value_counts(normalize=True).AIAN if 'AIAN' in df[self.race] else 1
        other_ratio = df[self.race].value_counts(normalize=True).OTHER if 'OTHER' in df[self.race] else 1
        white_ratio = df[self.race].value_counts(normalize=True).WHIATE if 'WHIATE' in df[self.race] else 1

        df["sample_weight"] = df[self.race].map(
            {
                "AAPI": (0.061 / 1.022) / aapi_ratio,
                "BLACK": (0.134 / 1.022) / black_ratio,
                "HISPANIC": (0.185 / 1.022) / hispanic_ratio,
                "AIAN": (0.013 / 1.022) / aian_ratio,
                "OTHER": (0.028 / 1.022) / other_ratio,
                "WHITE": (0.601 / 1.022) / white_ratio,
            }
        )

        # Split working data
        df.reset_index(inplace=True)
        X = df.copy()
        X.drop([self.race, "sample_weight"], axis=1, inplace=True)

        if self.geo_key == df.index.name:
            y = df[[self.geo_key, self.race, "sample_weight"]]
        else:
            y = df[[self.key, self.geo_key, self.race, "sample_weight"]]

        # Train (80) + Test(20)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.20, random_state=9)

        save_feather(X_train, self.outputs_path, f"X_train.feather")
        save_feather(y_train, self.outputs_path, f"y_train.feather")
        save_feather(X_test, self.outputs_path, f"X_test.feather")
        save_feather(y_test, self.outputs_path, f"y_test.feather")

        return (X_train, X_test, y_train, y_test)


class ZRP_Build(BaseZRP):
    """
    This class builds a new custom ZRP model trained off of user input data. Supply standard ZRP requirements including name, address, and race to build your custom model-pipeline. Race & ethnicity probablities and labels are returned from this class. The pipeline, model, and supporting data is saved automatically to "./artifacts/experiments/{zrp_model_name}/{zrp_model_source}/data/" in the support files path defined.
    
    Parameters
    ----------
    file_path: str
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model.
    zrp_model_source: str
        Indicates the source of zrp_modeling data to use. There are three optional sources 'bg' (for block_group), 'ct' (for census_tract), and 'zp' (for zip_code). By default 'census_tract' is inferred.
    """

    def __init__(self, file_path=None, zrp_model_name='zrp_0', zrp_model_source='ct', *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.zrp_model_name = zrp_model_name
        self.zrp_model_source = zrp_model_source
        self.outputs_path = os.path.join(self.out_path,
                                         "experiments",
                                         self.zrp_model_name,
                                         self.zrp_model_source,
                                         "data")
        self.geo_key = 'GEOID'

    def fit(self):
        return self

    def transform(self, data):
        make_directory(self.outputs_path)
        sample_path = self.outputs_path

        # Prepare data
        data = data.rename(columns = {self.first_name : "first_name", 
                              self.middle_name : "middle_name", 
                              self.last_name : "last_name",
                              self.house_number : "house_number", 
                              self.street_address : "street_address", 
                              self.city : "city",
                              self.zip_code : "zip_code",
                              self.state : "state", 
                              self.block_group : "block_group", 
                              self.census_tract : "census_tract"
                             }
                  )
        data = data.drop_duplicates(subset=['ZEST_KEY'])
        z_prepare = ZRP_Prepare(file_path=self.file_path)
        z_prepare.fit(data)
        prepared_data = z_prepare.transform(data)

        # Data Sampling 
        dsamp = ZRP_DataSampling(file_path=self.file_path)
        X_train, X_test, y_train, y_test = dsamp.transform(prepared_data)

        data = data.drop_duplicates(subset=['ZEST_KEY'])
        print("Post-sampling shape: ", data.shape)

        print("Unique labels: ", y_train['race'].unique())
        print("Other unique labels: ", y_test['race'].unique())
        cur_path = dirname(__file__)
        feature_list = load_json(os.path.join(cur_path, f'feature_list_{self.zrp_model_source}.json'))

        y_train = y_train.drop_duplicates(self.key)
        train_keys = list(y_train[self.key].values)
        X_train = X_train[X_train[self.key].isin(train_keys)]
        X_train = X_train.drop_duplicates(self.key)

        y_train[[self.geo_key, self.key]] = y_train[[self.geo_key, self.key]].astype(str)
        sample_weights = y_train[[self.key, 'sample_weight']].copy()

        assert X_train.shape[0] == y_train.shape[
            0], "Unexpected mismatch between shapes. There are duplicates in the data, please remove duplicates & resubmit the data"

        #### Set Index
        X_train.set_index(self.key, inplace=True)
        y_train.set_index(self.key, inplace=True)
        sample_weights.set_index(self.key, inplace=True)
        X_train.sort_index(inplace=True)
        y_train.sort_index(inplace=True)
        sample_weights.sort_index(inplace=True)

        feature_cols = list(set(X_train.columns) - set([self.key, self.geo_key, 'GEOID_BG', 'GEOID_CT',
                                                        'GEOID_ZIP', "first_name", "middle_name",
                                                        "last_name", 'ZEST_KEY_COL']))

        print('   train to numeric')
        X_train[feature_cols] = X_train[feature_cols].apply(pd.to_numeric, errors='coerce')

        print('\n---\nSaving raw data')
        save_feather(X_train, self.outputs_path, "train_raw_data.feather")
        save_feather(y_train, self.outputs_path, "train_raw_targets.feather")

        # Build Pipeline
        build_pipe = ZRP_Build_Pipeline(file_path=self.file_path)
        build_pipe.fit(X_train, y_train)
        X_train_fe = build_pipe.transform(X_train)

        # Build Model
        build_model = ZRP_Build_Model(file_path=self.file_path)
        build_model.fit(X_train_fe, y_train)
        y_hat_train, y_phat_train = build_model.transform(X_train_fe)

        pred_dict = {}
        pred_dict['labels'] = y_hat_train
        pred_dict['probablities'] = y_phat_train
        return (pred_dict)



from os.path import dirname, join, expanduser
from .preprocessing import *
from .base import BaseZRP
from .utils import *
import pandas as pd
import numpy as np
import statistics
import json
import sys
import os
import re

import warnings
warnings.filterwarnings(action='ignore')

def get_reduced(tmp_data):
    keep_cols = ['ZEST_KEY', 'first_name', 'middle_name', 'last_name',
                 'house_number', 'street_address', 'city', 'state', 'zip_code',
                 'BLKGRPCE', 'BLKGRPCE10', 'COUNTYFP', 'COUNTYFP10', 'FROMHN', 'TOHN',
                 'LFROMADD', 'LTOADD', 'PUMACE', 'PUMACE10', 'RFROMADD', 'RTOADD', 'SIDE',
                 'STATEFP', 'STATEFP10', 'TBLKGPCE', 'TRACTCE', 'TRACTCE10', 'TTRACTCE',
                 'ZCTA5CE', 'ZCTA5CE10', 'ZEST_FULLNAME', 'ZEST_KEY_COL', 'ZEST_STATE',
                 'ZEST_ZIP', 'GEOID_ZIP', 'GEOID_CT', 'GEOID_BG', 'age', 'original_ethnicity',
                 'original_race', 'original_sex', 'ethnicity', 'race', 'sex', 'source']
    na_match_cols = ['BLKGRPCE', 'BLKGRPCE10', 'COUNTYFP', 'COUNTYFP10', 'FROMHN', 'TOHN',
                     'LFROMADD', 'LTOADD', 'PUMACE', 'PUMACE10', 'RFROMADD', 'RTOADD', 'SIDE',
                     'STATEFP', 'STATEFP10', 'TBLKGPCE', 'TRACTCE', 'TRACTCE10', 'TTRACTCE',
                     'ZCTA5CE', 'ZCTA5CE10', 'ZEST_FULLNAME', 'ZEST_STATE', 'ZEST_ZIP']
    red_bit = keep_cols + ['HN_Match', 'ZIP_Match', 'RAW_ZEST_STATEFP']

    tmp_data = tmp_data.filter(red_bit)

    geocd = tmp_data.copy()
    nomatch = tmp_data.copy()

    geocd = geocd[(geocd.HN_Match.astype(float) == 1) & (geocd.ZIP_Match.astype(float) == 1)]
    geokeys = list(geocd['ZEST_KEY'].unique())

    nomatch = nomatch[~nomatch['ZEST_KEY'].isin(geokeys)]
    nomatch = nomatch.drop_duplicates('ZEST_KEY')

    geocd['TRACTCE'] = geocd.groupby('ZEST_KEY')['TRACTCE'].transform(lambda x: x.value_counts().idxmax())
    geocd['BLKGRPCE'] = geocd.groupby('ZEST_KEY')['BLKGRPCE'].transform(lambda x: x.value_counts().idxmax())
    geocd['ZCTA5CE'] = geocd.groupby('ZEST_KEY')['ZCTA5CE'].transform(lambda x: x.value_counts().idxmax())
    geocd['COUNTYFP'] = geocd.groupby('ZEST_KEY')['COUNTYFP'].transform(lambda x: x.value_counts().idxmax())

    geocd = geocd.drop_duplicates('ZEST_KEY')
    geocd["GEOID_CT"] = geocd[["RAW_ZEST_STATEFP", "COUNTYFP", "TRACTCE"]].apply(lambda x: "".join(x.dropna()), axis=1)
    geocd["GEOID_BG"] = geocd[["GEOID_CT", "BLKGRPCE"]].apply(lambda x: "".join(x.dropna()), axis=1)
    geocd = geocd.set_index('ZEST_KEY')

    if len(nomatch) > 1:
        nomatch[na_match_cols] = None
        nomatch = nomatch.set_index('ZEST_KEY')
        data_out = pd.concat([geocd, nomatch])
        data_out["GEOID_ZIP"] = np.where(data_out["ZCTA5CE"].isna(), data_out.zip_code, data_out["ZCTA5CE"])
        
        data_out = data_out.filter(keep_cols)
    else:
        data_out = geocd.filter(keep_cols)
        data_out["GEOID_ZIP"] = np.where(data_out["ZCTA5CE"].isna(), data_out.zip_code, data_out["ZCTA5CE"])
    
    data_out["GEOID"] = None
    data_out["GEOID"].fillna(data_out["GEOID_BG"]).fillna(data_out["GEOID_CT"]).fillna(data_out["GEOID_ZIP"])
    return(data_out)



def geo_search(geo_files_path, year, st_cty_code):
    """
    Returns a list of files associated with the state county code

    Parameters
    ----------
    geo_files_path:
        A string representing file path of the folder containing geo lookup tables
    year:
        A string year
    st_cty_code:
        A string for the state city code
    """
    file_list = []
    for root, dirs, files in os.walk(os.path.join(geo_files_path)):
        for file in files:
            if (st_cty_code in file):
                if year in file:
                    file_list.append(os.path.join(root, file))
    return (file_list)


def geo_read(file_list):
    """
    Returns a dataframe from files associated with the state county code

    Parameters
    ----------
    file_list:
        A list of strings representing file paths
    """
    aef = pd.DataFrame()
    for file in file_list:
        tmp = load_file(file)
        aef = pd.concat([aef, tmp], axis=0)
    return (aef)


def geo_zoom(geo_df):
    """
    Matches census tract
    
    Parameters
    ----------
    geo_df: pd.DataFrame
        Dataframe with geo data
    """
    geo_df = geo_df[(geo_df.HN_Match == 1) &
                    (geo_df.ZIP_Match == 1)]
    return (geo_df)


def geo_range(geo_df):
    """
    Define house number range indicators
    
    Parameters
    ----------
    geo_df: pd.DataFrame
        Dataframe with geo data
    """
    geo_df["small"] = np.where(
        (geo_df.FROMHN > geo_df.TOHN) & (geo_df.FROMHN.str.len() >= geo_df.TOHN.str.len()),
        geo_df.TOHN,
        geo_df.FROMHN)

    geo_df["big"] = np.where(
        (geo_df.FROMHN > geo_df.TOHN) & (geo_df.FROMHN.str.len() >= geo_df.TOHN.str.len()),
        geo_df.FROMHN,
        geo_df.TOHN)
    return(geo_df)    


class ZGeo(BaseZRP):
    """
    This class geocodes addresses.
    
    Parameters
    ----------
    file_path: str
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    """

    def __init__(self, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.key = 'ZEST_KEY'

    def fit(self):
        return self

    def geo_match(self, geo_df):
        """
        Returns match indicators

        Parameters
        ----------
        geo_df: pd.DataFrame
            Dataframe with geo data
        """
        geo_df["HN_Match"] = np.where(
            (geo_df[self.house_number] <= geo_df.big) &
            (geo_df[self.house_number] >= geo_df.small),
            1,
            0)
        
        geo_df["ZIP_Match_1"] = np.where(geo_df.ZEST_ZIP == geo_df[self.zip_code], 1, 0)
        geo_df["ZIP_Match_2"] = np.where(geo_df.ZCTA5CE10 == geo_df[self.zip_code], 1, 0)   
        geo_df["NEW_SUPER_ZIP"] = np.where(geo_df.ZIP_Match_1 == 1, geo_df.ZEST_ZIP, geo_df.ZCTA5CE10)
        geo_df["ZIP_Match"] = np.where(geo_df.NEW_SUPER_ZIP == geo_df[self.zip_code], 1, 0)
        
        
        return(geo_df)    
    
    def transform(self, input_data, geo, processed, replicate, save_table=True):
        """
        Returns a DataFrame of geocoded addresses.

        :param input_data: A pd.DataFrame.
        :param geo: A String
        :param processed: A boolean.
        :param replicate: A boolean.
        :param save_table: A boolean. Tables are saved if True. Default is True.
        :return: A DataFrame
        """
        curpath = dirname(__file__)
        out_geo_path = os.path.join(curpath, '../data/processed/geo/2019')

        print("")
        # Load Data
        try:
            data = input_data.copy()
            print("   Data is loaded")
        except AttributeError:
            data = load_file(self.file_path)
            print("   Data file is loaded")
        
        data = data.rename(columns = {self.first_name : "first_name", 
                                      self.middle_name : "middle_name", 
                                      self.last_name : "last_name",
                                      self.house_number : "house_number", 
                                      self.street_address : "street_address", 
                                      self.city : "city",
                                      self.zip_code : "zip_code",
                                      self.state : "state", 
                                      self.block_group : "block_group", 
                                      self.census_tract : "census_tract"
                             }
                  )
            
        prg = ProcessGeo()
        data = prg.transform(data, processed=processed, replicate=replicate)
        
        print("   [Start] Mapping geo data")
        state = most_common(list(data[self.state].unique()))
        
        if len(geo)>2:
            file_list = geo_search(out_geo_path, self.year, geo)
            aef = geo_read(file_list)
        if len(geo) <= 2:
            aef = load_file(os.path.join(out_geo_path, f"Zest_Geo_Lookup_{self.year}_State_{geo}.parquet"))

        data["ZEST_FULLNAME"] = data[self.street_address]
        print("      ...merge user input & lookup table")
        geo_df = aef.merge(data, on="ZEST_FULLNAME", how="right")
        
        geo_df = geo_range(geo_df)

        geo_df = self.geo_match(geo_df)
        
        print("      ...mapping")
        all_keys = list(geo_df[self.key].unique())
        odf = geo_df.copy()
        geo_df = geo_zoom(geo_df)
        da_zoom = geo_df.copy()
        
        geocoded_keys = list(geo_df[self.key].unique())
        add_na_keys = list(set(all_keys) - set(geocoded_keys))
        odf = odf[odf[self.key].isin(add_na_keys)]

        geo_df = pd.concat([geo_df, odf])

        geo_df = get_reduced(geo_df)

        if save_table:
            make_directory(self.out_path)
            if self.runname is not None:
                file_name = f"Zest_Geocoded_{self.runname}_{self.year}__{geo}.parquet"
            else:
                file_name = f"Zest_Geocoded__{self.year}__{geo}.parquet"
            save_dataframe(geo_df, self.out_path, file_name)
        print("   [Completed] Mapping geo data")
        return (geo_df)



from .acs_scaler import CustomRatios
from .app_fe import AppFeatureEngineering, NameAggregation
from .app_preprocessor import HandleCompoundNames
from .set_key import SetKey


__all__ = ['CustomRatios', 'AppFeatureEngineering', 'NameAggregation', 'HandleCompoundNames', 'SetKey']


__title__ = "zrp"
__version__ = "0.1.19"
__release_download_url_prefix__ = "https://github.com/zestai/zrp/archive/refs/tags/"
__download_url_prefix__ = "https://github.com/zestai/zrp/releases/download/"
__projects__ = "https://github.com/zestai"
__projects_branch__ = "main"





from .predict import PredictPass, ZRP_Predict_ZipCode, ZRP_Predict_BlockGroup, ZRP_Predict_CensusTract, ZRP_Predict, BISGWrapper, FEtoPredict
from .performance import ZRP_Performance
from .pipeline_builder import ZRP_Build_Pipeline, ZRP_Build_Model, ZRP_DataSampling, ZRP_Build


__all__ = ['PredictPass', 'ZRP_Predict_ZipCode', 'ZRP_Predict_BlockGroup', 
           'ZRP_Predict_CensusTract', 'ZRP_Predict', 'FEtoPredict', 'ZRP_Performance',
          'ZRP_Build_Pipeline', 'ZRP_Build_Model', 'ZRP_DataSampling', 'ZRP_Build']



import pandas as pd
import numpy as np
import os
import re
import sys
from os.path import join, expanduser
import json
import fiona
import warnings
warnings.filterwarnings("ignore")
    

def load_json(path):
    """
    Load json files
    
    Parameters
    ----------
    path: str
        Filepath of json file
    """
    with open(path, "r") as infile:
        data = json.load(infile)
    return data


def save_json(data, path, file_name):
    """
    Save json files
    
    Parameters
    ----------
    data: list or dict
        Data to save to json
    path: str
        Filepath of where to store json
    file_name: str
        Name of file
    """    
    with open(os.path.join(path, file_name), "w") as outfile:
        json.dump(data, outfile)


def save_dataframe(data, path, file_name):
    """
    Save dataframe as parquet
    
    Parameters
    ----------
    data: pd.DataFrame
        Dataframe to save
    path: str
        Filepath of where to store json
    file_name: str
        Name of file
    """      
    data.to_parquet(os.path.join(path, file_name))
    return (print("Output saved"))


def save_feather(data, path, file_name):
    """
    Save dataframe as feather file
    
    Parameters
    ----------
    data: pd.DataFrame
        Dataframe to save
    path: str
        Filepath of where to store json
    file_name: str
        Name of file
    """          
    data.reset_index(drop = False).to_feather(os.path.join(path,
                                                         file_name))
    return(print("Output saved"))
        
def make_directory(output_directory = None):
    """
    Creates a directory
    
    Parameters
    ----------
    output_directory: str
        Filepath or name of directory to create
    """      

    try:
        if output_directory == None:
            os.makedirs("artifacts")
        else:
            os.makedirs(os.path.join(output_directory))
    except FileExistsError:
        print("Directory already exists")
        pass

def load_file(file_path):    
    """
    Load files. Compatible with csv, text, feather, xlsx, and parquet
    
    Parameters
    ----------
    file_path: str
        File path of file to load
    """

    na_values = ["None",
                 "NAN",
                 "NONE",
                 " ",
                 "(X)",
                 "-",
                 "  ",
                 "   ",
                 "-666666666",
                 "-999999999",
                 "-888888888"
                 ]
    if file_path.endswith(".csv"):
        data = pd.read_csv(file_path,
                           dtype=str,
                           na_values=na_values)
    elif file_path.endswith(".feather"):
        data = pd.read_feather(file_path)
        data = data.astype(str)
    elif file_path.endswith(".xlsx"):
        data = pd.read_excel(file_path,
                             dtype=str,
                             na_values=na_values)
    elif file_path.endswith(".parquet"):
        data = pd.read_parquet(file_path)
        data = data.astype(str)
    elif file_path.endswith(".txt"):
        with open(file_path) as f:
            first_line = f.readline()
            if "|" in first_line:
                data = pd.read_csv(file_path,
                                   sep="|",
                                   dtype=str,
                                   na_values=na_values)
            elif "," in first_line:
                data = pd.read_csv(file_path,
                                   sep=",",
                                   dtype=str,
                                   na_values=na_values)
            else:
                data = pd.read_csv(file_path,
                                   sep="\t",
                                   dtype=str,
                                   na_values=na_values)
    else:
        raise ValueError("Unrecognizable table format")
    return (data)


def load_mappings(support_files_path):
    """
    Loads support mapping files
    
    Parameters
    ----------
    support_files_path: str
        Filepath or name of directory where files are stored
    """        
    mapping_file_path = os.path.expanduser(support_files_path)
    print(mapping_file_path)
    # add zip to county mapping
    state_mapping = load_json(os.path.join(mapping_file_path,
                                           "state_mapping.json"))
    street_suffix_mapping = load_json(os.path.join(mapping_file_path,
                                                   "street_suffix_mapping.json"))
    directionals_mapping = load_json(os.path.join(mapping_file_path,
                                                  "directionals_mapping.json"))
    unit_mapping = load_json(os.path.join(mapping_file_path,
                                          "unit_mapping.json"))
    return (state_mapping,
            street_suffix_mapping,
            directionals_mapping,
            unit_mapping)


def gdbToDf_short(file, indx):
    """
    Loads shapefiles
    
    Parameters
    ----------
    file: str
        Filepath or name of file to load
    indx: str
        Indicates type of shapefile to load in
    """      
    dictList = []
    if indx == "address":
        with fiona.open(file) as src:
            for i in range(10000000):
                dictList.append(next(src)["properties"])
    else:
        with fiona.open(file) as src:
            for i in range(len(src)):
                dictList.append(next(src)["properties"])
    df = pd.DataFrame(
        dictList,
        columns=dictList[0].keys(),
    )
    return df


def acs_rename(data):
    """
    Rename ACS index column
    
    Parameters
    -----------
    data: pd.DataFrame
        Dataframe to make changes to
    """
    data_cols = list(data.columns[1:])
    data.columns = ["result"] + data_cols
    return data


def acs_trt_split(data, feature):
    """
    Extract keys from ACS identifier column for census tract level data
    
    Parameters
    ----------
    data: pd.DataFrame
        Dataframe to make changes to
    feature: str
        Name of identifier column to extract information from
    """
    data = data[data[feature].notna()]
    state = [
        re.split(f".*state:", re.split(r">", data[feature].iloc[i])[0])[1]
        for i in range(len(data))
    ]

    county = [
        re.split(f".*county:", re.split(r">", data[feature].iloc[i])[1])[1]
        for i in range(len(data))
    ]

    tract = [
        re.split(f".*tract:", re.split(r">", data[feature].iloc[i])[2])[1]
        for i in range(len(data))
    ]

    data["STATEFP"] = state
    data["COUNTYFP"] = county
    data["TRACTCE"] = tract
    data["GEO_KEY"] = data["STATEFP"] + data["COUNTYFP"] + data["TRACTCE"]
    if "TRACTCE10" in data.columns:
        data["GEO_KEY_10"] = data["STATEFP"] + data["COUNTYFP"] + data["TRACTCE10"]
        data = data.drop("TRACTCE10", axis=1)
    data = data.drop([feature, "STATEFP", "COUNTYFP", "TRACTCE"], axis=1)

    return data


def acs_zip_split(dataframe, feature):
    """
    Extract keys from ACS identifier column for zip code level data
    
    Parameters
    ----------
    data: pd.DataFrame
        Dataframe to make changes to
    feature: str
        Name of identifier column to extract information from
    """    
    state = [
        re.split(f".*state:", re.split(r">", dataframe[feature].iloc[i])[0])[1]
        for i in range(len(dataframe))
    ]
    zcta5 = [
        re.split(f".*zip code tabulation area:", dataframe["result"].iloc[i])[1]
        for i in range(len(dataframe))
    ]
    dataframe["STATEFP"] = state
    dataframe["ZEST_ZIP"] = zcta5
    dataframe = dataframe.drop(
        ["Unnamed: 0", "result"],
        axis=1
    )


def most_common(lizt):
    """
    Returns the most common element of a list or series
    
    Parameters
    ----------
    lizt: list
        List to extract most common element from
    """
    return max(set(lizt), key=lizt.count)



from .base import BaseZRP
from .acs_mapper import ACSModelPrep
from .geo_geocoder import ZGeo
from .preprocessing import  ProcessStrings, ProcessGeo, ProcessACS
from .prepare import ZRP_Prepare

__all__ = ['BaseZRP','ZRP_Prepare', 'ProcessStrings', 'ProcessGeo', 'ProcessACS', 'ACSModelPrep' ]


import numpy as np
import pandas as pd
import re
from sklearn.base import BaseEstimator, TransformerMixin 


class CustomRatios(BaseEstimator, TransformerMixin):
        """
    CustomRatios is used to convert the counts into percentages and ratios for ACS data, where possible. 
    """
        def __init__(self):
            self.house_val_cols = ['B25075_002',
                                    'B25075_003',
                                    'B25075_004',
                                    'B25075_005',
                                    'B25075_006',
                                    'B25075_007',
                                    'B25075_008',
                                    'B25075_009',
                                    'B25075_010',
                                    'B25075_011',
                                    'B25075_012',
                                    'B25075_013',
                                    'B25075_014',
                                    'B25075_015',
                                    'B25075_016',
                                    'B25075_017',
                                    'B25075_018',
                                    'B25075_019',
                                    'B25075_020',
                                    'B25075_021',
                                    'B25075_022',
                                    'B25075_023',
                                    'B25075_024',
                                    'B25075_025',
                                    'B25075_026',
                                    'B25075_027'
                                  ]
            self.income_cols = [    'B19001_002',
                                    'B19001_003',
                                    'B19001_004',
                                    'B19001_005',
                                    'B19001_006',
                                    'B19001_007',
                                    'B19001_008',
                                    'B19001_009',
                                    'B19001_010',
                                    'B19001_011',
                                    'B19001_012',
                                    'B19001_013',
                                    'B19001_014',
                                    'B19001_015',
                                    'B19001_016',
                                    'B19001_017']
            self.education_cols = ['B06009_001',
                                   'B06009_002',
                                        'B06009_003',
                                        'B06009_004',
                                        'B06009_005',
                                        'B06009_006']
            self.race_cols = ['B02001_001',
                              'B02001_002', 
                                'B02001_003', 
                                'B02001_004', 
                                'B02001_005', 
                                'B02001_006', 
                                'B02001_007', 
                                'B02001_008', 
                                'B02001_009', 
                                'B02001_010']
            self.transit_cols = [    'B08301_002',
                                    'B08301_003',
                                    'B08301_004',
                                    'B08301_011', 
                                    'B08301_012', 
                                    'B08301_013', 
                                    'B08301_016', 
                                    'B08301_017', 
                                    'B08301_018',
                                    'B08301_019',
                                    'B08301_020', 
                                    'B08301_021']
            self.hispanic_cols = ["B03001_001",
                                  "B03001_002", 
                                  "B03001_003",
                                 "B03001_008",
                                  "B03001_016"]
            self.single_ancestry_cols = ["B04004_001", 
                                  "B04004_006", 
                                  "B04004_035", 
                                  "B04004_038", 
                                  "B04004_073",
                                  "B04004_094"]
            self.reporting_ancestry_cols = ["B04006_001", 
                                  "B04006_006", 
                                  "B04006_035", 
                                  "B04006_038", 
                                  "B04006_073",
                                  "B04006_094"]
            self.ancestry_cols = ["B04007_001",
                                 "B04007_002",
                                 "B04007_005"]
            self.nativity_cols = ["B05012_001",
                                  "B05012_002",
                                  "B05012_003"]
            self.language_cols = ["C16001_001",
                                  "C16001_002",
                                  "C16001_003",
                                  "C16001_006",
                                  "C16001_009",
                                  "C16001_012",  
                                  "C16001_015", 
                                  "C16001_018",  
                                  "C16001_021",
                                  "C16001_024",  
                                  "C16001_029",  
                                  "C16001_030", 
                                  "C16001_033",  
                                  "C16001_036"]
            self.naturalization_cols = ["B05011_001",
                                        "B05011_002",
                                        "B05011_003",
                                        "B05011_004",
                                        "B05011_005", 
                                        "B05011_006",   
                                        "B05011_007", 
                                        "B05011_008",  
                                        "B05011_009",
                                        "B05011_010"]
            
        def fit(self, X, y):
            self.data_cols = list(X.columns)
            self.house_val_cols = list(set(self.house_val_cols).intersection(set(self.data_cols)))
            self.income_cols = list(set(self.income_cols).intersection(set(self.data_cols)))
            self.education_cols = list(set(self.education_cols).intersection(set(self.data_cols)))
            self.race_cols = list(set(self.race_cols).intersection(set(self.data_cols)))
            self.transit_cols = list(set(self.transit_cols).intersection(set(self.data_cols)))
            self.hispanic_cols = list(set(self.hispanic_cols).intersection(set(self.data_cols)))
            self.single_ancestry_cols = list(set(self.single_ancestry_cols).intersection(set(self.data_cols)))
            self.reporting_ancestry_cols = list(set(self.reporting_ancestry_cols).intersection(set(self.data_cols)))
            self.ancestry_cols = list(set(self.ancestry_cols).intersection(set(self.data_cols)))
            self.nativity_cols = list(set(self.nativity_cols).intersection(set(self.data_cols)))
            self.language_cols = list(set(self.language_cols).intersection(set(self.data_cols)))
            self.naturalization_cols = list(set(self.naturalization_cols).intersection(set(self.data_cols)))
            return self
 
                                  
        def transform(self, X):
            data= X.copy()
            
            if 'B99021_001' in self.data_cols:
                data['allocated_race'] = round(data['B99021_002'] / data ['B99021_001'], 6)
                data['unallocated_race'] = round(data['B99021_003'] / data ['B99021_001'], 6)

            # work 
            if ('B23020_001' in self.data_cols)  & ('B23020_003' in self.data_cols)  & ('B23020_002' in self.data_cols):
                data['weekly_work_ratio'] =  data['B23020_001']/40
                data['weekly_work_ratio_m'] =  data['B23020_002']/40
                data['weekly_work_ratio_w'] =  data['B23020_003']/40
                data['weekly_work_ratio_wm'] = round(data['B23020_003']/data['B23020_002'], 2)
                data['weekly_work_sum'] = data['B23020_003'] + data['B23020_002']

            # house value 
            for col in self.house_val_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5)
                
            # income
            for col in self.income_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 4)
                
            # education 
            for col in self.education_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 4)
            # hispanic_cols  
            for col in self.hispanic_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5)
                
            # single_ancestry_cols
            for col in self.single_ancestry_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5)
                
            # reporting_ancestry_cols 
            for col in self.reporting_ancestry_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5)
            # ancestry_cols  
            for col in self.ancestry_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 4)
                
            # nativity_cols
            for col in self.nativity_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5)
                
            # language_cols 
            for col in self.language_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5)                
            # naturalization 
            for col in self.naturalization_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5)    

            # race 
            for col in self.race_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'], 5) 
                
            data['race_minority_pct'] =  round((data['B02001_003']+ data['B02001_004']+ data['B02001_005']+ data['B02001_006']+ data['B02001_008']) / data['B02001_001'])
            
            data['race_minority_red_pct'] =  round((data['B02001_003']+ data['B02001_004']+ data['B02001_006']) / data['B02001_001'], 6)
            
            data['race_min'] =  round(data['B02001_003']+ data['B02001_004']+ data['B02001_006'], 6)
            
            data['race_minority_allocation_pct'] =  round((data['B02001_003']+ data['B02001_004']+ data['B02001_005']+ data['B02001_006']+ data['B02001_008'])/ data['B99021_001'], 6)
            data['race_aaw'] = round(data['B02001_005']/data['B02001_002'], 6)
            
            data['race_minw'] = round((data['B02001_003']+ data['B02001_004']+ data['B02001_005']+ data['B02001_006']+ data['B02001_008'])/data['B02001_002'], 4)
            
            data['race_aian'] = round(data['B02001_004']/ data["B02001_001"], 6)
            # add language, ethinicity, origin
            if ('B03001_003' in self.data_cols) & ('B03001_001' in self.data_cols):
                data['ethnicity_hispanic'] = round(data['B03001_003']/data['B03001_001'], 6)
            if 'B05012_003' in self.data_cols:
                data['pct_multinational'] = round(data['B05012_003']/data['B05012_001'], 4)
            if 'C16001_030' in self.data_cols:
                data['pct_aapi_language'] = round((data["C16001_018"] + data["C16001_021"] + data["C16001_024"] + data["C16001_029"] + data["C16001_030"]) / data["C16001_001"], 6)
            if 'C16001_003' in self.data_cols:
                data['pct_spanish'] = round(data["C16001_003"] / data["C16001_001"], 4)
            # transit
            for col in self.transit_cols:
                data[col+'_pct'] = round(data[col]/ data['B01003_001'],5) 
                
            raw_counts = list(self.house_val_cols + self.income_cols +  self.education_cols +  self.transit_cols  +  self.single_ancestry_cols +  self.ancestry_cols +  self.nativity_cols +  self.naturalization_cols + self.reporting_ancestry_cols + self.language_cols)
            raw_counts_list = list(set(data.columns).intersection(set(raw_counts)))
            data = data.drop(raw_counts_list, axis=1)
            data = data.replace([np.inf, -np.inf], np.nan)
            return(data)
                

            



import numpy as np
import pandas as pd
import re
from sklearn.base import BaseEstimator, TransformerMixin 

class HandleCompoundNames(BaseEstimator, TransformerMixin):
    '''This class handles compound surnames.
    
    Parameters
    ----------
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    geo_key: str
        Name of Census GEOID column
    race: str
        Name of race column    
    ''' 
    

    def __init__(self, last_name, first_name, middle_name):
        self.last_name = last_name
        self.first_name = first_name
        self.middle_name = middle_name
        
    def fit(self, X, y):
        return self
    
    def _upper_case(self, X, columns):
        """Convert strings in the Series to uppercase.
        
        Parameters
        ----------
        X: dataframe
        columns: list
        """
        for col in columns:
            X[col] = X[col].str.upper()
        return X

    def _handle_compounds(self, X):
        """
        Simple compund surname handler. An individual with n names will receive a sample weight 1/n (where n = number of names).
        
        Parameters
        ----------
        X: dataframe
        """
        X[self.last_name] = X[self.last_name].str.replace('-', ' ', regex=False) 
        X[self.last_name] = X[self.last_name].str.replace(' +', ' ', regex=True)

        compound_name_str_all = X[self.last_name].str.split(' ', expand=True) # split compounds from non_compounds
        non_compound = X[compound_name_str_all[1].isna()].copy().reset_index(drop=True)
        compound = X[~compound_name_str_all[1].isna()].copy().reset_index(drop=True)

       
        compound_name_str = compound[self.last_name].str.split(' ', expand=True)
        n_compounds = compound_name_str.shape[1] # max number of unique strings
        compound_name_int = compound_name_str.copy() # binary representation of string
        for col in compound_name_str.columns:
            compound_name_int[col] = (~compound_name_str[col].isna()).astype(int)
        compound_name_int['compound_count'] = np.sum(compound_name_int, axis=1) #  row-wise count of strings  

        compound_tiled = compound.loc[compound.index.repeat(n_compounds)].reset_index(drop=True)
        sw_tiled = compound_name_int.loc[compound_name_int.index.repeat(n_compounds)].reset_index(drop=True)
        unrolled_names = (compound_name_str.values.ravel())
        compound_tiled[self.last_name] = unrolled_names
        compound_result = compound_tiled[~compound_tiled[self.last_name].isna()]

        joint_result = non_compound.append(compound_result).reset_index(drop=True)

        return joint_result

    def transform(self, X):
        data = X.copy()
        data.reset_index(drop=False, inplace=True)
        # compound names (row indicies are not preserved!)
        data = self._handle_compounds(data)
        return(data)


from zrp.modeling.predict import BISGWrapper, ZRP_Predict
from os.path import dirname, join, expanduser
from zrp.prepare.prepare import ZRP_Prepare
from zrp.prepare.base import BaseZRP
from zrp.prepare.utils import *
import pandas as pd
import numpy as np
import warnings
import surgeo
import pickle
import joblib
import json
import pycm
import sys
import os
import re


class ZRP(BaseZRP):
    """
    Zest Race Predictor, predicts race & ethnicity using name & geograhpic data
    
    Parameters
    ----------
    support_files_path: "data/processed"
        File path with support data.
    key: str, default 'ZEST_KEY'
        Key to set as index. If not provided, a key will be generated.
    first_name: str, default 'first_name'
        Name of first name column
    middle_name: str, default 'middle_name'
        Name of middle name column
    last_name: str, default 'last_name'
        Name of last name/surname column
    house_number: str, default 'house_number'
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str, default 'street_address'
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str, default 'city'
        Name of city column.
    state: str, default 'state'
        Name of state column.
    zip_code: str, default 'zip_code'
        Name of zip or postal code column.
    race: str, default 'race'
        Name of the race column
    proxy: str, default 'probs'
        Specifies whether to return proxy results as 'probs' (the proxy probabilities) or 'labels'.
    census_tract: str
        Name of census tract column
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list, optional
        List of missing values to replace.
    file_path: str, optional
        Path where to put artifacts and other files generated during intermediate steps. 
    geocode: bool, default True
        Whether to geocode.
    bisg: bool, default True
        Whether to return BISG proxies. 
    readout: bool, default True
        Whether to return a readout.
    n_jobs: int, default 49
        Number of jobs in parallel
    year: str, default '2019'
        ACS year to use.
    span: str, default '5'
        Year span of ACS data to use.
    runname: str, default 'test'
    """

    def __init__(self, file_path=None, pipe_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.pipe_path = pipe_path

    def fit(self):
        return self

    def transform(self, input_data):
        """
        Processes input data and generates ZRP predictions. Generates BISG predictions additionally if specified.

        Parameters
        -----------
        input_data: pd.Dataframe
            Dataframe to be transformed
        """
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.file_path)
            
        data = data.rename(columns = {self.first_name : "first_name", 
                              self.middle_name : "middle_name", 
                              self.last_name : "last_name",
                              self.house_number : "house_number", 
                              self.street_address : "street_address", 
                              self.city : "city",
                              self.zip_code : "zip_code",
                              self.state : "state", 
                              self.block_group : "block_group", 
                              self.census_tract : "census_tract"
                             }
                  )
        
        make_directory(self.out_path)

        z_prepare = ZRP_Prepare(file_path=self.file_path)
        z_prepare.fit(data)
        prepared_data = z_prepare.transform(data)

        curpath = dirname(__file__)
        if self.pipe_path is None:
            self.pipe_path = join(curpath, "modeling/models")

        z_predict = ZRP_Predict(file_path=self.file_path, pipe_path=self.pipe_path)
        z_predict.fit(prepared_data)
        predict_out = z_predict.transform(prepared_data)

        if self.bisg:
            bisg_data = prepared_data.copy()
            bisg_data = bisg_data[~bisg_data.index.duplicated(keep='first')]

            bisgw = BISGWrapper()
            full_bisg_proxies = bisgw.transform(bisg_data)
            save_feather(full_bisg_proxies, self.out_path, f"bisg_proxy_output.feather")

        try:
            predict_out = input_data.merge(predict_out.reset_index(drop=False), on=self.key)
        except KeyError:
            pass

        return (predict_out)



from os.path import dirname, join, expanduser
from joblib import Parallel, delayed
from .base import BaseZRP
import multiprocessing
from tqdm import tqdm
from .utils import *
import pandas as pd
import numpy as np 
import json
import sys
import os
import re
from zrp.validate import *


def norm_na(data, na_values):
    """
    Standardize missing values
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    na_values: list
        List of missing values to replace
    """
    base_na_list = ["\\b#N/A\\b",
                    "\\b#N/A N/A\\b",
                    "\\b#NA\\b",
                    "\\b-1.#IND\\b",
                    "\\b-1.#QNAN\\b",
                    "\\b-NAN\\b",
                    "\\b-NaN\\b",
                    "\\b-nan\\b",
                    "\\b1.#IND\\b",
                    "\\b1.#QNAN\\b",
                    "\\b<NA>\\b",
                    "\\bN/A\\b",
                    "\\bNA\\b",
                    "\\bNULL\\b",
                    "\\bNAN\\b",
                    "\\bNaN\\b",
                    "\\bn/a\\b",
                    "\\bnan\\b",
                    "\\bnull\\b",
                    "\\bNULL\\b",
                    "\\bNONE\\b",
                    "^//(X//)$",
                    "^-$",
                   "^\\s*$"] 

    if na_values:
        word_border = "\\b"
        na_values = [word_border + s + word_border for s in na_values]
        na_values = na_values + base_na_list
    else:
        na_values = base_na_list

    na_dict = {}    
    for key in na_values:
        na_dict[key] = None

    data = data.replace(na_dict, regex=True)
    return(data)


def set_id(data, key):
    """
    Set Key
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    key: str
        Key to set as index
    data_cols: list
        List of data columns available
    """
    if key == data.index.name:
        print("The key is already set")
    elif key in list(data.columns):
        data = data.set_index(key)
    else:
        data["tmp_key"] = data[list(data.columns)].apply(lambda row:\
                                                "_".join(row.values.astype(str)),
                                                axis = 1)
        if data["tmp_key"].nunique() == len(data):
            data = data.rename(columns = {"tmp_key": key})
            data = data.set_index(key)
        else:
            data = data.sort_values("tmp_key")
            data[key] = data["tmp_key"] + data.index.astype(str)
            data = data.set_index(key)
    return(data)


def reduce_whitespace(data):
    """
    Reduce whitespace
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    """
    return data.apply(lambda x: x.str.strip().str.replace(" +", " ", regex=True))


def replicate_address(data, i, street_suffix_mapping, unit_mapping):
    """
    Replicate street addresses 
    
    Parameters
    ----------
    data: pd.DataFrame
       ACS dataFrame to make changes to 
    street_address: str
       Name of street address column 
    street_suffix_mapping: dict
       Dictionary with street mappings
    unit_mapping: dict
       Dictionary with building unit mapping   
    """
    df_base =  data[i]     # base is complete, containing the original record (1)

    data[i] = pd.DataFrame([{i: data[i]}]).replace(unit_mapping, regex=True).loc[0, i]
    df_u = data[i].strip()
    data[i] = pd.DataFrame([{i: data[i]}]).replace(street_suffix_mapping, regex=True).loc[0, i] # Second Unit & Street Suffix abbreviation mapping
    
    df_sus =  data[i]     # street & unit mapping + split after suffix (2)
    df_suu =  data[i]     # street & unit mapping + split before unit (3)
    df_sus = df_sus.split("-", 1)[0].strip()
    df_u = df_u.split("-", 1)[0]
    df_suu = df_suu.split("-", 1)[0].strip()
    data[i] = data[i].replace("-", "").strip()
 
    df_su =  data[i]      # suffix & unit mapped to abbreviations (4) 
    df_no =  data[i]      # remove all numbers not-attached to str (5)

    df_no = re.sub("^([0-9]{1,6}[A-Z]{1,3})", "", df_no)
    df_no = re.sub(" [0-9]{4,7} ", "", df_no)

    dataout = pd.DataFrame([df_base,
                           df_u,
                           df_su,
                           df_sus,
                           df_suu,
                           df_no
                          ])
    dataout.index = [i, f"{i}_u", f"{i}_su", f"{i}_sus", f"{i}_suu",f"{i}_no" ]
    dataout = dataout.drop_duplicates()
    return(dataout)


def address_mining(data, i):
    """
    Cleans street addresses

    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    i: int
        Iteration variable
    """
    data[i]  = re.sub("[^A-Za-z0-9\\s]",
                                   "",
                                   re.sub("[^A-Za-z0-9']",
                                          " ",
                                          str(data[i])))
    return(data[i])


class HandleTracts():
    """
    HandleTracts parses the American Community Survey response format into state, county, and tracts.
    It will also create a GEO_KEY for merging with geocoded user data.
    """
    def __init__(self):
        super().__init__()
    
    def fit(self, data):
        pass
    
    def transform(self, data):
        """
        Parameters
        ----------
        data: pd.DataFrame
           ACS dataFrame to make changes to 
        """
        if not isinstance(data, pd.DataFrame):
            raise ValueError("Input must be a dataframe")
        acs_ct = data.copy()
        acs_zip = data.copy()
        
        acs_rename(acs_ct)
        acs_rename(acs_zip)
        
        acs_trt_split(acs_ct, feature="result")
        acs_zip_split(acs_zip, feature="result")
        
        return(acs_ct, acs_zip)

    
def replicate_address_2(data, street_address, street_suffix_mapping, unit_mapping):
    """
    Replicate street addresses 
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    street_address: str
       Name of street address column 
    street_suffix_mapping: dict
       Dictionary with street mappings
    unit_mapping: dict
       Dictionary with building unit mapping   
    """
    # base
    data = data.reset_index(drop=False)
    print("         ...Base")
    df_base =  data.copy()# base is complete, containing the original record (1)
    print("         ...Map street suffixes...")
    data[street_address] = data[street_address].replace(street_suffix_mapping, regex=True) # this mapping takes the longest but is ok for 10K records
    data[street_address] = np.where(data[street_address]=='nan', None, data[street_address])
    print("         ...Mapped & split by street suffixes...")
    # Remove addtl "-"
    data[street_address] = data[street_address].str.split("-", 1, expand=True)[0]
    data[street_address] = data[street_address].str.replace(pat = "-", repl = "", regex=False)
    print("         ...Number processing...")
    data[street_address] = data[street_address].str.replace("^([0-9]{1,6}[A-Z]{1,3})", "", regex=True)
    data[street_address] = data[street_address].str.replace("^([0-9]{1,3} [A-Z]{2})", "", regex=True)
    data[street_address] = data[street_address].str.replace(" [0-9]{4,7} ", "", regex=True)
    print("")
    dataout = pd.concat([df_base,
                           data
                          ], axis=0)
    dataout = dataout.drop_duplicates()
    print(f"         Address dataframe expansion is complete! (n={len(dataout)})")
    return(dataout)


def replicate_house_number(data, house_number):
    """
    Replicate street addresses 
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    house_number: str
       Name of street address column 
    """
    # base
    data = data.reset_index(drop=True)
    print("         ...Base")
    df_base =  data.copy()# base is complete, containing the original record (1)
    print("         ...Number processing...")
    data[house_number].apply(lambda x: re.sub("[^0-9]",\
                                                     "",\
                                                    str(x))) 
    dataout = pd.concat([df_base,
                           data
                          ], axis=0)
    dataout = dataout.drop_duplicates()
    print(f"         House number dataframe expansion is complete! (n={len(dataout)})")
    return(dataout)

       
class ProcessStrings(BaseZRP):
    """
    ProcessStrings executes all ZRP preprocessing. All user data is processed with additional  processing operations for geo-specific and American Community Survey data.
    
    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    support_files_path:
        File path with support data
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list
        List of missing values to replace 
    file_path: str
        Input data file path
    geocode: bool
        Whether to geocode
    race: str
        Name of race column
    bisg: bool, default True
        Whether to return BISG proxies
    readout: bool
        Whether to return a readout
    n_jobs: int (default 1)
        Number of jobs in parallel
    """
        
    def __init__(self, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        
            
    def fit(self, data):
        data_cols = list(data.columns)
        print("   [Start] Validating input data")
        if self.census_tract in data_cols:
            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.census_tract]
        elif (self.census_tract in data_cols) & (self.block_group in data_cols):
            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.census_tract, self.block_group]
        elif (self.zip_code in data_cols) & (self.geocode==True):
            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.zip_code, self.house_number, self.street_address, self.city, self.state]
        elif (self.zip_code in data_cols) & (self.geocode==False):
            self.required_cols = [self.first_name, self.middle_name, self.last_name, self.zip_code]
        val_na = is_missing(data, self.required_cols)
        if val_na:
            assert True, f"     Missing required data {val_na}"
        validate = ValidateInput()
        validate.fit()
        validators_in = validate.transform(data)
        save_json(validators_in, self.out_path, "input_validator.json")
        print("   [Completed] Validating input data")
        print("")
        return self

    
    def transform(self, data_in, replicate=False):
        """
        Processes general data

        Parameters
        ----------
        data_in: pd.DataFrame
            DataFrame to make changes to 
        replicate: bool
            Indicator to process address components
        """
        
        # Load Data
        try:
            data = data_in.copy()
        except AttributeError:
            data = load_file(self.file_path)
        
        data_cols =  data.columns
        data = set_id(data, self.key)
        
        numeric_cols =  list(set([self.zip_code,
                                  self.census_tract,
                                  self.house_number
                                 ]).intersection(set(data_cols)))

        # Convert to uppercase & trim whitespace
        print("   Formatting P1")
        data = data.astype(str)
        data = data.apply(lambda x: x.str.upper())
        
        data = reduce_whitespace(data)
        na_dict =  {"^\\s*$": None,
                    "^NAN$": None,
                    "^NONE$": None}
        data = data.replace(na_dict, regex=True)     
        
        # Remove/replace special characters
        for col in numeric_cols:
            data[col] = data[col].apply(lambda x: re.sub("[^0-9]",\
                                                     "",\
                                                    str(x))) 
        if self.last_name:
            name_cols = list(set([self.first_name,
                                  self.middle_name,
                                  self.last_name]).intersection(set(data_cols))) 
            for col in name_cols:
                    data[col] =  data[col].apply(lambda x: re.sub("[^A-Za-z\\s]",
                                                              "",
                                                              re.sub("[^A-Za-z']",
                                                                     " ",
                                                                     str(x))))
        print("   Formatting P2")
        print("   reduce whitespace")
        data = reduce_whitespace(data)
        return(data)
    
        
class  ProcessACS(BaseZRP):
    """
    All user data is processed with additional processing operations for geo-specific and American Community Survey data.

    Parameters
    ----------
    out_path: str
        Path to store output artifacts
    """    
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
        
            
    def fit(self, data):
        data_cols = list(data.columns)
        print("   [Start] Validating input ACS data")
        validate = ValidateGeocoded()
        validate.fit()
        acs_validator = validate.transform(data)
        save_json(acs_validator, self.out_path, "input_acs_validator.json")
        print("   [Completed] Validating input ACS data")
        print("")
        
        return self


    def transform(self, data):
        """
        Processes ACS data 

        Parameters
        ----------
        data_in: pd.DataFrame
            DataFrame to make changes to 
        """
        data = reduce_whitespace(data)
        data = data.replace({'\\bN\\b': None}, regex=True) # Note: 800+ columns
        data = data.reset_index(drop=False)
        return(data)
    
    
class  ProcessGeo(BaseZRP):
    """
    All user data is processed with additional  processing operations for geo-specific and American Community Survey data.

    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    support_files_path:
        File path with support data
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list
        List of missing values to replace 
    file_path: str
        Input data file path
    geocode: bool
        Whether to geocode
    race: str
        Name of race column
    bisg: bool, default True
        Whether to return BISG proxies
    readout: bool
        Whether to return a readout
    n_jobs: int (default 1)
        Number of jobs in parallel
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
            
    def fit(self, data):
        data_cols = list(data.columns)
        print("   [Start] Validating input geo data")
        if self.census_tract in data_cols:
            self.required_cols = [self.census_tract]
        elif (self.census_tract in data_cols) & (self.block_group in data_cols):
            self.required_cols = [self.census_tract, self.block_group]
        elif (self.zip_code in data_cols) & (self.geocode==True):
            self.required_cols = [self.zip_code, self.house_number, self.street_address, self.city, self.state]
        val_na = is_missing(data, self.required_cols)
        val_na = is_missing(data, self.required_cols)
        if val_na:
            assert True, f"Missing required data {val_na}"
        geo_validate = ValidateGeo()
        geo_validate.fit()
        geo_validators_in = geo_validate.transform(data)
        save_json(geo_validators_in, self.out_path, "input_geo_validator.json") 
        print("   [Completed] Validating input geo data")
        
        return self


    def transform(self, data, processed, replicate=False):
        """
        Processes geo data 

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame to make changes to 
        processed: bool
            Indicator is the data was previously cleaned & processed 
        replicate: bool
            Indicator to process compenents of the address
        """        
        curpath = dirname(__file__)
        print("   [Start] Processing geo data")
        # Load Data
        if not processed:
            data_cols =  data.columns
            data = set_id(data, self.key)
            
            numeric_cols =  list(set([self.zip_code,
                                      self.census_tract,
                                      self.house_number
                                     ]).intersection(set(data_cols)))
            

            print("      ...formatting")
            data = data.astype(str)
            data = data.apply(lambda x: x.str.upper())
            
            data = reduce_whitespace(data)
            print("")
            print(data.state.unique())
            
            na_dict =  {"^\\s*$": None,
                       "^NAN$": None,
                       "^NONE$": None}
            data = data.replace(na_dict, regex=True)
            print("")
            print(data.state.unique())
            # Remove/replace special characters
            for col in numeric_cols:
                data[col] = data[col].apply(lambda x: re.sub("[^0-9]",\
                                                         "",\
                                                        str(x))) 
            if self.last_name:
                name_cols = list(set([self.first_name,
                                      self.middle_name,
                                      self.last_name]).intersection(set(data_cols))) 
                for col in name_cols:
                        data[col] =  data[col].apply(lambda x: re.sub("[^A-Za-z\\s]",
                                                                  "",
                                                                  re.sub("[^A-Za-z']",
                                                                         " ",
                                                                         str(x))))
        
        data_path = join(curpath, f'../data/processed')
        state_mapping, street_suffix_mapping, directionals_mapping, unit_mapping = load_mappings(data_path)
        
        if self.key in data.columns:
            data["ZEST_KEY_COL"] = data[self.key]
        else:
            data["ZEST_KEY_COL"] = data.index        
        
        print("      ...address cleaning")
        street_addr_dict = dict(zip(data.index, data[self.street_address])) 
        street_addr_results = Parallel(n_jobs = self.n_jobs, prefer="threads", verbose=1)(delayed((address_mining))(street_addr_dict, i) for i in tqdm(list(data.index)))

        data[self.street_address] = street_addr_results
        data[self.city]  = data[self.city].str.replace("[^\\w\\s]", "", regex=True)

        # State
        data[self.state]  = data[self.state].str.replace("[^\\w\\s]", "", regex=True)
        data[self.state] = data[self.state].replace(state_mapping)
        # Zip code 
        data[self.zip_code] = np.where((data[self.zip_code].isna()) |\
                                 (data[self.zip_code].str.contains("None")),
                                  None,
                                  data[self.zip_code].apply(lambda x: x.zfill(5)))
        data[self.zip_code] = data[self.zip_code].astype(str).str[:5]        
        
        if replicate:
            print("      ...replicating address")
            data = replicate_address_2(data, self.street_address, street_suffix_mapping, unit_mapping)
            data = replicate_house_number(data, self.house_number)
        
        print("      ...formatting")
        addr_cols = list(set(list(data.columns)).intersection(set([self.zip_code, self.census_tract, self.house_number, self.city, self.state, self.street_address])))


        data = reduce_whitespace(data)
        print("   [Completed] Processing geo data")
        return(data)

    
class  ProcessGLookUp(BaseZRP):
    """
    All user data is processed with additional  processing operations for geo-specific and American Community Survey data.
    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    support_files_path:
        File path with support data
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list
        List of missing values to replace 
    file_path: str
        Input data file path
    geocode: bool
        Whether to geocode
    race: str
        Name of race column
    bisg: bool, default True
        Whether to return BISG proxies
    readout: bool
        Whether to return a readout
    n_jobs: int (default 1)
        Number of jobs in parallel
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.street_address= "ZEST_FULLNAME"
        self.state = "STATEFP"
        self.zip_code = "ZEST_ZIP"
        self.census_tract = "TRACTCE" 
        self.block_group = "BLKGRPCE"
        self.county = "COUNTYFP"
        
            
    def fit(self):
        return self


    def transform(self, data, state_mapping):
        """
        Processes geo lookup table data 

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame to make changes to 
        state_mapping: dict
            Mapping of state codes and names to standard abbreviations 
        """               
        print("   [Start] Processing lookup data")
        # Load Data
        data_cols =  data.columns
        numeric_cols =  list(set([self.zip_code,
                                  self.census_tract,
                                  self.block_group,
                                  self.county,
                                 ]).intersection(set(data_cols)))
        data = data.astype(str)
        data = data.apply(lambda x: x.str.upper())
        
        data = reduce_whitespace(data)
        na_dict =  {"^\\s*$": None,
                    "^NAN$": None,
                    "^NONE$": None} 
        data = data.replace(na_dict, regex=True)
        
        # Remove/replace special characters
        for col in numeric_cols:
            data[col] = data[col].apply(lambda x: re.sub("[^0-9]",\
                                                     "",\
                                                    str(x))) 

      
        data[self.census_tract] = np.where((data[self.census_tract].isna()) |\
                                 (data[self.census_tract].str.contains("None")),
                                  None,
                                  data[self.census_tract].apply(lambda x: x.zfill(6)))
        data[self.block_group] = np.where((data[self.block_group].isna()) |\
                                 (data[self.block_group].str.contains("None")),
                                  None,
                                  data[self.block_group])
        data[self.county] = np.where((data[self.county].isna()) |\
                                 (data[self.county].str.contains("None")),
                                  None,
                                  data[self.county].apply(lambda x: x.zfill(3))) 
        
        print("     ...processing")
        spec_cols = list(set(list(data_cols)).intersection(set([self.zip_code, self.block_group, self.county, self.state, self.census_tract])))
        data[spec_cols] = norm_na(data[spec_cols], self.na_values)
        data = reduce_whitespace(data)
        print("   [Completed] Processing lookup data")
        
        data_cols = list(data.columns)
        validate = ValidateInput()
        validate.fit()
        validator_in = validate.transform(data) 
        print(validator_in)
        return(data)



import shutil
import zipfile
import os
import sys
import warnings
from urllib.request import urlretrieve
from tqdm import tqdm
from zrp import about


# This is used to show progress when downloading.
# see here: https://github.com/tqdm/tqdm#hooks-and-callbacks
class TqdmUpTo(tqdm):
    """Provides `update_to(n)` which uses `tqdm.update(delta_n)`."""
    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b  : int, optional
            Number of blocks transferred so far [default: 1].
        bsize  : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize  : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def download_progress(url, fname):
    """
    Download a file and show a progress bar.
    :param url: A string for the url of the release zip to download.
    :param fname: A string for the local file name under which the downloaded file can be found.
    :return:
    """

    with TqdmUpTo(unit='B', unit_scale=True, miniters=1,
              desc=url.split('/')[-1]) as t:  # all optional kwargs

        print(f"Retrieving url at {url}")
        urlretrieve(url, filename=fname, reporthook=t.update_to, data=None)
        t.total = t.n
    return fname


def download_and_clean_lookup_tables(url, lookup_tables_output_fname, lookup_tables_output_zip_fname, geo_yr="2019",
                                     acs_yr="2019", acs_range="5yr"):
    """
    Downloads look up tables and file them within the module.
    This downloads the zip file from the repository, extracts it, renames it, then moves the
    tables to the correct directory, and removes large files not used at runtime.
    :param lookup_tables_output_fname: A string for the name of the file downloaded after unzipping.
    :param acs_range: A string for the year range the acs lookup table data will be from.
    :param acs_yr: A string for the year the acs lookup table data will be from.
    :param geo_yr: A string for the year the geo lookup table data will be from.
    :param url: A string for the url of the release zip to download.
    :param lookup_tables_output_zip_fname: A string for the name of the zip file downloaded.
    :return:
    """
    cwd = os.path.dirname(os.path.abspath(__file__))
    path_to_lt_zip = os.path.join(cwd, lookup_tables_output_zip_fname)
    print("Downloading zrp release...", file=sys.stderr)
    download_progress(url, path_to_lt_zip)
    print("Finished download.")
    print("\n")

    print("Filing extras...")
    with zipfile.ZipFile(path_to_lt_zip, 'r') as zf:
        zf.extractall(cwd)
    os.remove(path_to_lt_zip)

    # Get rid of prefix that unzipping prepends
    # curr_folder = cwd.split("/")[-1]
    # unzipped_src_fname = curr_folder + "-" + lookup_tables_output_fname
    # path_to_unzipped_src = os.path.join(cwd, unzipped_src_fname)
    path_to_lookup_tables = os.path.join(cwd, lookup_tables_output_fname)
    # os.rename(path_to_unzipped_src, path_to_lookup_tables)

    # Clear old look up table directories
    data_dir = os.path.join(cwd, 'data')
    geo_data_dir = os.path.join(data_dir, f'processed/geo/{geo_yr}')
    acs_data_dir = os.path.join(data_dir, f'processed/acs/{acs_yr}/{acs_range}')

    if os.path.isdir(geo_data_dir):
        shutil.rmtree(geo_data_dir)
    if os.path.isdir(acs_data_dir):
        shutil.rmtree(acs_data_dir)

    print("Old geo lookup table data cleared out.")

    # Migrate lookup tables
    dl_geo_dir = os.path.join(cwd, lookup_tables_output_fname, f'geo/{geo_yr}')
    dl_acs_dir = os.path.join(cwd, lookup_tables_output_fname, f'acs/{acs_yr}/{acs_range}')

    if os.path.isdir(dl_geo_dir):
        shutil.move(dl_geo_dir, geo_data_dir)
        print("New geo lookup tables successfully migrated.")
    else:
        warnings.warn(f"The geo lookup data was not found in {dl_geo_dir}. Ensure you're requesting a valid year. "
                      "Consult the lookup_tables release to troubleshoot.")
    if os.path.isdir(dl_acs_dir):
        shutil.move(dl_acs_dir, acs_data_dir)
    else:
        warnings.warn(f"The acs lookup data was not found in {dl_acs_dir}. Ensure you're requesting a valid year and"
                      "year range. Consult the lookup_tables release to troubleshoot.")

    # Remove rest of lookup table folder
    shutil.rmtree(path_to_lookup_tables)

    # save a version file so we can tell what it is
    vpath = os.path.join(data_dir, 'version')
    with open(vpath, 'w') as vfile:
        vfile.write('zrp release --> {}'.format(lookup_tables_output_fname))

    print("Filed lookup tables successfully.", file=sys.stderr)


def download_and_clean_pipelines(url, pipelines_output_fname, pipelines_output_zip_fname):
    """
    Downloads pipeline pickle files and file them within the module.
    This downloads the zip file from the repository, extracts it, renames it, then moves the
    tables to the correct directory, and removes large files not used at runtime.
    :param pipeline_output_fname: A string for the name of the file downloaded after unzipping.
    :param url: A string for the url of the release zip to download.
    :param pipelines_output_zip_fname: A string for the name of the zip file downloaded.
    :return:
    """
    cwd = os.path.dirname(os.path.abspath(__file__))
    path_to_ppln_zip = os.path.join(cwd, pipelines_output_zip_fname)
    print("Downloading zrp release...", file=sys.stderr)
    download_progress(url, path_to_ppln_zip)
    print("Finished download.")
    print("\n")

    print("Filing extras...")
    with zipfile.ZipFile(path_to_ppln_zip, 'r') as zf:
        zf.extractall(cwd)
    os.remove(path_to_ppln_zip)

    # Get rid of prefix that unzipping prepends
    # curr_folder = cwd.split("/")[-1]
    # unzipped_src_fname = curr_folder + "-" + pipelines_output_fname
    # path_to_unzipped_src = os.path.join(cwd, unzipped_src_fname)
    path_to_pipelines = os.path.join(cwd, pipelines_output_fname)
    # os.rename(path_to_unzipped_src, path_to_pipelines)

    # Clear old look up table directories
    model_dir = os.path.join(cwd, 'modeling/models')
    block_group_dir = os.path.join(model_dir, 'block_group')
    census_tract_dir = os.path.join(model_dir, 'census_tract')
    zip_code_dir = os.path.join(model_dir, 'zip_code')

    block_group_pipeline = os.path.join(block_group_dir, 'pipe.pkl')
    census_tract_pipeline = os.path.join(census_tract_dir, 'pipe.pkl')
    zip_code_pipeline = os.path.join(zip_code_dir, 'pipe.pkl')

    if os.path.isfile(block_group_pipeline):
        os.remove(block_group_pipeline)
    if os.path.isfile(census_tract_pipeline):
        os.remove(census_tract_pipeline)
    if os.path.isfile(zip_code_pipeline):
        os.remove(zip_code_pipeline)
    print("Old pipelines cleared out.")

    # Migrate pipelines
    dl_bg_pipe_file = os.path.join(path_to_pipelines, 'block_group_pipe.pkl')
    dl_ct_pipe_file = os.path.join(path_to_pipelines, 'census_tract_pipe.pkl')
    dl_zp_pipe_file = os.path.join(path_to_pipelines, 'zip_code_pipe.pkl')

    if os.path.isfile(dl_bg_pipe_file):
        shutil.move(dl_bg_pipe_file, os.path.join(block_group_dir, 'pipe.pkl'))
        print("Block group pipeline successfully migrated.")
    else:
        warnings.warn(f"The block group pipeline was not found in {dl_bg_pipe_file}."
                      "Consult the pipelines release to troubleshoot.")

    if os.path.isfile(dl_ct_pipe_file):
        shutil.move(dl_ct_pipe_file, os.path.join(census_tract_dir, 'pipe.pkl'))
        print("Census tract pipeline successfully migrated.")
    else:
        warnings.warn(f"The census tract pipeline was not found in {dl_ct_pipe_file}."
                      "Consult the pipelines release to troubleshoot.")

    if os.path.isfile(dl_zp_pipe_file):
        shutil.move(dl_zp_pipe_file, os.path.join(zip_code_dir, 'pipe.pkl'))
        print("Zip code pipeline successfully migrated.")
    else:
        warnings.warn(f"The zip code pipeline was not found in {dl_zp_pipe_file}."
                      "Consult the pipelines release to troubleshoot.")

    # Remove rest of pipelines folder
    shutil.rmtree(path_to_pipelines)

    # save a version file so we can tell what it is
    data_dir = os.path.join(cwd, 'data')
    vpath = os.path.join(data_dir, 'version')
    with open(vpath, 'w') as vfile:
        vfile.write('zrp release --> {}'.format(pipelines_output_fname))

    print("Filed pipelines successfully.", file=sys.stderr)


def get_release():
    version = about.__version__
    dl_tpl = "{m}-{v}"
    return dl_tpl.format(m="zrp", v=version)


def download():
    release_pkg = get_release()

    # lookup_tables_output_fname = release_pkg + "_lookup_tables"
    lookup_tables_output_fname = "lookup_tables"
    lookup_tables_output_zip_fname = release_pkg + "_lookup_tables" + ".zip"
    lookup_table_url = about.__download_url_prefix__ + release_pkg + "/lookup_tables.zip"
    download_and_clean_lookup_tables(lookup_table_url, lookup_tables_output_fname, lookup_tables_output_zip_fname)

    pipelines_output_fname = "pipelines"
    pipelines_output_zip_fname = release_pkg + "_pipelines" + ".zip"
    pipelines_url = about.__download_url_prefix__ + release_pkg + "/pipelines.zip"
    download_and_clean_pipelines(pipelines_url, pipelines_output_fname, pipelines_output_zip_fname)



from os.path import dirname, join, expanduser
from zrp.prepare.preprocessing import *
from zrp.prepare.geo_geocoder import *
from zrp.prepare.acs_mapper import *
from zrp.prepare.base import BaseZRP
from zrp.prepare.utils import *
from zrp.validate import *
import pandas as pd
import numpy as np
import warnings
import glob
import json
import tqdm
import sys
import os
import re
import pickle
import joblib
import surgeo
import xgboost



from zrp.modeling import src
from zrp.modeling.src.app_preprocessor import HandleCompoundNames
from zrp.modeling.src.acs_scaler import CustomRatios
from zrp.modeling.src.app_fe import AppFeatureEngineering, NameAggregation


class PredictPass(BaseZRP):
    """
    Generates proxies

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data):
        """
        Processes input data and generates BISG predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
        proxies = pd.DataFrame({"AAPI": None, "AIAN": None, "BLACK": None,
                                "HISPANIC": None, "WHITE": None, f"{self.race}_proxy": None}, index = data.index)
        return(proxies) 

def validate_case(data, key, last_name):
    df = data.copy()
    new_row = df.tail(1).reset_index(drop=False)
    new_row[key] = "validate_case_001"
    new_row[last_name] = "SMITH JONES"
    new_row = new_row.set_index(key)
    df = pd.concat([df, new_row])
    return(df)
    
def validate_drop(data):
    data = data.drop('validate_case_001')
    return(data) 


class BISGWrapper(BaseZRP):
    """
    Wrapper function for Bayesian Improved Surname Geocoding

    Generates proxies using BISG algorithm.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        
    def fit(self, input_data):
        """

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """        
        if (self.last_name not in input_data.columns):
            raise ValueError('Last name needs to be provided when initializing this class. Please provide last name data in a column named "last_name" or set the custom name of the last name column in the data')
        if self.zip_code not in input_data.columns:
            raise ValueError('Zip or postal code name needs to be provided when initializing this class. Please provide zip code data in a column named "zip_ode" or set the custom name of the zip code column in the data')
        return self
    
    def transform(self, input_data):
        """
        Processes input data and generates BISG predictions.

        Parameters
        -----------
        data: pd.Dataframe
            Dataframe to be transformed
        """
        df = input_data.copy()
        df = df[~df.index.duplicated(keep='first')]

        df = df.filter([self.last_name, self.zip_code, self.census_tract])
        df[self.zip_code] = np.where((df[self.zip_code] == "") | (df[self.zip_code] == " ") | (df[self.zip_code] == None) | (df[self.zip_code].str.len()<5), "99999", df[self.zip_code])
        
        bisg = surgeo.SurgeoModel()
        bisg_results = bisg.get_probabilities(names  = df[self.last_name].reset_index(drop=True),
                                              geo_df = df[self.zip_code].astype(int))
        combo = df.reset_index().merge(bisg_results, 
                           how="left", 
                           left_on=[self.last_name, self.zip_code], 
                           right_on=["name", "zcta5"]
                          )
        combo.drop(columns = ['zcta5', 'name'], inplace=True)
        
        combo.rename(columns = {
            'white': 'WHITE',
            'black': 'BLACK',
            'api': 'AAPI',
            'native': 'AIAN',
            'multiple': 'OTHER',
            'hispanic': 'HISPANIC'
        }, inplace=True) 
        combo = combo.set_index(self.key)
        # Generate proxy at threshold
        subset = combo[['WHITE', 'BLACK', 'AAPI', 'AIAN', 'HISPANIC'
                       ]]
        identifiedRaces = subset.idxmax(axis=1)
        combo[f"{self.race}_proxy"] = identifiedRaces
        combo['source_bisg'] = 1
        proxies = combo[["AAPI", "AIAN", "BLACK", "HISPANIC",
                         "WHITE", f"{self.race}_proxy", "source_bisg"]]            
        return(proxies)



class ZRP_Predict_ZipCode(BaseZRP):
    """
    Generates proxies using model trained on zip code features.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """

        src_path = os.path.join(self.pipe_path, "zip_code")
        sys.path.append(src_path)        
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
            
        numeric_cols = list(data.filter(regex='^B|^C16').columns)
        data[numeric_cols] = data[numeric_cols].apply(pd.to_numeric, errors='coerce') 
        
        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))
#         model = pd.read_pickle(os.path.join(src_path,"model.pkl") )
        pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl") )
        
        
        data = validate_case(data, self.key, self.last_name)
        fe_data = pipe.transform(data)
        fe_data = validate_drop(fe_data)
        fe_matrix = xgboost.DMatrix(fe_data)
        
        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
        proxies.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        proxies['source_zip_code'] = 1
        return(proxies)
        


class ZRP_Predict_BlockGroup(BaseZRP):
    """
    Generates proxies using model trained on block group features.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """        

        src_path = os.path.join(self.pipe_path,"block_group")
        sys.path.append(src_path)
        
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
            
        numeric_cols = list(data.filter(regex='^B|^C16').columns)
        data[numeric_cols] = data[numeric_cols].apply(pd.to_numeric, errors='coerce')
        
        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))
#         model = pd.read_pickle(os.path.join(src_path,"model.pkl") )
        pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl") )
        
        data = validate_case(data, self.key, self.last_name)
        fe_data = pipe.transform(data)
        fe_data = validate_drop(fe_data)
        fe_matrix = xgboost.DMatrix(fe_data)
        
        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
        proxies.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        proxies['source_block_group'] = 1
        return(proxies)
        



class ZRP_Predict_CensusTract(BaseZRP):
    """
    Generates proxies using model trained on census tract features.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    """

    def __init__(self, pipe_path, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        
    def fit(self):
        return self
    
    def transform(self, input_data):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """        

        src_path = os.path.join(self.pipe_path,"census_tract")
        sys.path.append(src_path)
        
        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
            
        numeric_cols = list(data.filter(regex='^B|^C16').columns)
        data[numeric_cols] = data[numeric_cols].apply(pd.to_numeric, errors='coerce') 

        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))        
#         model = pd.read_pickle(os.path.join(src_path,"model.pkl") )
        pipe = pd.read_pickle(os.path.join(src_path, "pipe.pkl") )
        
        data = validate_case(data, self.key, self.last_name)
        fe_data = pipe.transform(data)
        fe_data = validate_drop(fe_data)
        fe_matrix = xgboost.DMatrix(fe_data)
        
        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
        proxies.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        proxies['source_census_tract'] = 1
        return(proxies)
    


class ZRP_Predict(BaseZRP):
    """
    Generates race proxies.
    Attempts to predict on census tract, then block group, then zip code based on which level ACS data is found for. If
    Geo level data is unattainable, the BISG proxy is computed. No prediction returned if BISG cannot be computed either.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    file_path: str
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    """

    def __init__(self, pipe_path, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.pipe_path = pipe_path
        self.census_tract = 'GEOID_CT'
        self.block_group = 'GEOID_BG'
        self.zip_code = 'GEOID_ZIP'
        
        
    def fit(self, data):
        assert xgboost.__version__ == "1.0.2", "XGBoost version does not match requirements, required version is 1.0.2"  
        data_cols =  list(data.columns)
        self.required_cols = [self.first_name, self.middle_name, self.last_name, "GEOID", "B01003_001"]
        val_na = is_missing(data, self.required_cols)
        if val_na:
            assert True, f"Missing required data {val_na}"        

        print("   [Start] Validating pipeline input data")
        validator = ValidateInput()
        validator.fit()
        validator_in = validator.transform(data)
        save_json(validator_in, self.out_path, "input_predict_validator.json")
        print("   [Completed] Validating pipeline input data")
        print("")        
        return self
    
    def transform(self, input_data, save_table=True):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """        

        # Load Data
        try:
            data = input_data.copy()
        except AttributeError:
            data = load_file(self.proxy_data)
        cur_path = dirname(__file__)

        flb = load_json(f'{cur_path}/feature_list_bg.json')
        flc = load_json(f'{cur_path}/feature_list_ct.json')
        flz = load_json(f'{cur_path}/feature_list_zp.json')

        if 'acs_source' in data.columns:    
            df_0 = data[data['acs_source'] == 'BG']
            keys_0 = list(df_0.index.values)
            df_1 = data[~(data.index.isin(keys_0)) 
                         & (data['acs_source'] == 'CT')]
            keys_1 = keys_0 + list(df_1.index.values)
            df_2 = data[~(data.index.isin(keys_1)) 
                          & (data['acs_source'] == 'ZIP')]
            keys_2 = keys_1 + list(df_2.index.values)
            df_3 = data[~(data.index.isin(keys_2))
                                & (data['acs_source'].isna())]
        else:
            raise KeyError("Processed data is required for ZRP_Predict. Please use EnginetoPredict if supplying feature engineered data to generate predictions" )
            
        out_list = []
        if not df_0.empty:
            zrp_bg = ZRP_Predict_BlockGroup(self.pipe_path)
            out_0 = zrp_bg.transform(df_0.filter(flb))
            out_list.append(out_0)    
        if not df_1.empty:
            zrp_ct = ZRP_Predict_CensusTract(self.pipe_path)
            out_1 = zrp_ct.transform(df_1.filter(flc))
            out_list.append(out_1)    
        if not df_2.empty:
            zrp_zp = ZRP_Predict_ZipCode(self.pipe_path)
            out_2 = zrp_zp.transform(df_2.filter(flz))
            out_list.append(out_2)    
        if not df_3.empty:
            bisgw = BISGWrapper()
            bisgw.fit(df_3)
            out_3 = bisgw.transform(df_3)
            out_list.append(out_3)  
            
        proxies_out = pd.concat(out_list)

        source_cols = list(set(proxies_out.columns).intersection(set([
            'source_block_group', 'source_census_tract',
            'source_zip_code', 'source_bisg'])))
        proxies_out[source_cols] = proxies_out[source_cols].fillna(0)
        proxies_out = proxies_out.sort_values(source_cols)
        
        if save_table:
            make_directory(self.out_path)
            file_name = f"proxy_output.feather"
            save_feather(proxies_out,
                         self.out_path,
                         file_name)      
        
        return(proxies_out)

    
    
class FEtoPredict(BaseZRP):
    """
    Generates proxies from feature engineered data.

    Parameters
    ----------
    pipe_path: str
        Folder path to directory containing pipeline
    pipe_type: str, (default='census_tract')
        Type of pipeline that generated the engineered data.
        Options: 'block_group', 'census_tract', or 'zip_code'
    """

    def __init__(self, pipe_path, pipe_type='census_tract', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipe_path = pipe_path
        self.pipe_type = pipe_type
        
        
    def fit(self):
        return self
    
    def transform(self, input_data, save_table=True):
        """
        Processes input data and generates ZRP proxy predictions.

        Parameters
        -----------
        input_data: pd.DataFrame
            Dataframe to be transformed
        """
        model = xgboost.Booster()
        model.load_model(os.path.join(src_path,"model.txt"))
#         model = pd.read_pickle(os.path.join(self.pipe_path, f"{pipe_type}/model.pkl") )
        # Load Data
        assert not input_data.empty, "Feature engineered data is empty or missing. Please provide the feature engineered data as `input_data` to generate predictions."
        fe_data = input_data.copy()
        fe_matrix = xgboost.DMatrix(fe_data)

        proxies = pd.DataFrame(model.predict(fe_matrix), index = fe_data.index)
        proxies.columns = ["AAPI", "AIAN", "BLACK", "HISPANIC", "WHITE"]
        proxies[f"{self.race}_proxy"] = proxies.idxmax(axis=1)
        proxies[f'source_{pipe_type}'] = 1        

        
        if save_table:
            make_directory(self.out_path)
            file_name = f"{self.pipe_type}_proxy_output.feather"
            save_feather(proxies,
                         self.out_path,
                         file_name)
            
        return(proxies)
    
